"""Data object storing the solver parameters for the Nexus Simulation runs."""

from dataclasses import dataclass

from ResSimpy.Enums.TimeSteppingMethodEnum import TimeSteppingMethod
from ResSimpy.DataModelBaseClasses.SolverParameter import SolverParameter
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping


@dataclass(kw_only=True)
class NexusSolverParameter(SolverParameter):
    """Data object storing the solver parameters for the Nexus Simulation runs.

    Nexus is time-dependent for the solver parameters.
    """

    # Time Stepping options
    timestepping_method: TimeSteppingMethod | None = None
    dt_auto: float | None = None
    dt_min: float | None = None
    dt_max: float | None = None
    dt_max_increase: float | None = None
    dt_con: float | None = None
    dt_vipts: float | None = None
    dt_qmaxprod: float | None = None
    dt_qmaxinj: float | None = None
    dt_connopen: float | None = None
    dt_maxincafcut: float | None = None
    dt_adjusttotime: float | None = None
    dt_reduceafcut: float | None = None
    dt_wcycle: float | None = None
    dt_gcycle: float | None = None
    dt_vip_maxincrease: float | None = None
    dt_vip_maxincafcut: float | None = None

    # Solver Options for grid/specific reservoir/all
    solver_reservoir_cycle_length: float | None = None
    solver_reservoir_max_cycles: float | None = None
    solver_reservoir_globaltol: float | None = None
    solver_reservoir_equation_solver: str | None = None
    solver_global_cycle_length: float | None = None
    solver_global_max_cycles: float | None = None
    solver_global_globaltol: float | None = None
    solver_global_equation_solver: str | None = None
    solver_all_cycle_length: float | None = None
    solver_all_max_cycles: float | None = None
    solver_all_globaltol: float | None = None
    solver_all_equation_solver: str | None = None

    # Independent Solver options and settings
    solver_timestep_cut: bool | None = None
    solver_precon: str | None = None
    solver_precon_setting: str | None = None
    solver_precon_value: float | None = None
    solver_facilities: str | None = None
    solver_ksub_method: str | None = None
    solver_dual_solver: bool | None = None  # Default is ON
    solver_system_reduced: bool | None = None  # Default is ON
    solver_nbad: float | None = None
    solver_pressure_coupling: str | None = None
    solver_pseudo_slack: bool | None = None
    solver_mumps_solver: str | None = None

    # Material balance options
    implicit_mbal: str | None = None

    # Implicit stability (IMPSTAB) options
    impstab_on: bool | None = None  # Default is OFF
    impstab_criteria: str | None = None
    impstab_skip_mass_cfl: bool | None = None
    impstab_target_cfl: float | None = None
    impstab_limit_cfl: float | None = None
    impstab_no_cuts: float | None = None
    impstab_max_cuts: float | None = None
    impstab_skip_block_dcmax: float | None = None

    # Gridsolver options
    gridsolver_implicit_coupling_setting: str | None = None
    gridsolver_impes_coupling_setting: str | None = None
    gridsolver_implicit_precon_setting: str | None = None
    gridsolver_cpr_pressure_equation: str | None = None
    gridsolver_press_reduction: float | None = None
    gridsolver_implicit_reduction: float | None = None
    gridsolver_grid_reduction: float | None = None

    # PERFREV options
    perfrev: str | None = None

    # maximum change and other settings for the simulation
    maxnewtons: float | None = None
    maxbadnets: float | None = None
    cutfactor: float | None = None
    negmasscut: float | None = None
    dvollim: float | None = None
    dzlim: float | None = None
    dslim: float | None = None
    dplim: float | None = None
    dmoblim: float | None = None
    dsglim: float | None = None
    negflowlim: float | None = None
    negmassaqu: float | None = None
    krdamp: float | None = None
    volerr_prev: float | None = None
    sgctol: float | None = None
    egsgtol: float | None = None
    sgcperftol: float | None = None
    line_search: float | None = None
    perfp_damp: float | None = None

    # tolerances
    tols_volcon: float | None = None
    tols_mass: float | None = None
    tols_target: float | None = None
    tols_wellmbal: float | None = None
    tols_perf: float | None = None

    # settings for implicit, impes grids.
    dcmax_implicit: float | None = None
    dcmax_impes: float | None = None
    dcmax_all: float | None = None

    dcrpt_implicit: float | None = None
    dcrpt_impes: float | None = None
    dcrpt_all: float | None = None

    volrpt_implicit: float | None = None
    volrpt_impes: float | None = None
    volrpt_all: float | None = None

    dcrpt_vip_implicit: float | None = None
    dcrpt_vip_impes: float | None = None
    dcrpt_vip_all: float | None = None

    dzmax_vip_implicit: float | None = None
    dzmax_vip_impes: float | None = None
    dzmax_vip_all: float | None = None

    dpmax_vip_implicit: float | None = None
    dpmax_vip_impes: float | None = None
    dpmax_vip_all: float | None = None

    dvmax_vip_implicit: float | None = None
    dvmax_vip_impes: float | None = None
    dvmax_vip_all: float | None = None

    dsmax_vip_implicit: float | None = None
    dsmax_vip_impes: float | None = None
    dsmax_vip_all: float | None = None

    def _write_out_solver_param_block(self) -> None:
        """Writes out the solver parameter block for an instance of NexusSolverParameter."""
        raise NotImplementedError

    @staticmethod
    def dt_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the keyword mapping from simulator keyword to ResSimpy attribute and the type of the object.

        Associated with the DT keywords in Nexus. e.g. of the form "DT MAX 50".
        """
        # DT keywords
        dt_keyword_map: dict[str, tuple[str, type]] = {

            'AUTO': ('dt_auto', float),
            'MIN': ('dt_min', float),
            'MAX': ('dt_max', float),
            'MAXINCREASE': ('dt_max_increase', float),
            'CON': ('dt_con', float),
            'VIPTS': ('dt_vipts', float),
            'QMAXPROD': ('dt_qmaxprod', float),
            'QMAXINJ': ('dt_qmaxinj', float),
            'CONNOPEN': ('dt_connopen', float),
            'MAXINCAFCUT': ('dt_maxincafcut', float),
            'ADJUSTTOTIME': ('dt_adjusttotime', float),
            'REDUCEAFCUT': ('dt_reduceafcut', float),
            'WCYCLE': ('dt_wcycle', float),
            'GCYCLE': ('dt_gcycle', float),
            'VIP_MAXINCREASE': ('dt_vip_maxincrease', float),
            'VIP_MAXINCAFCUT': ('dt_vip_maxincafcut', float),
        }
        return dt_keyword_map

    @staticmethod
    def solver_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the keyword mapping from simulator keyword to ResSimpy attribute and the type of the object.

        Associated with the SOLVER keywords in Nexus. e.g. of the form "SOLVER ALL CYCLELENGTH 10".
        """
        # Solver keywords
        solver_keyword_map = {
            'ALL CYCLELENGTH': ('solver_all_cycle_length', float),
            'ALL MAXCYCLES': ('solver_all_max_cycles', float),
            'ALL GLOBALTOL': ('solver_all_globaltol', float),
            'ALL ITERATIVE': ('solver_all_equation_solver', str),
            'ALL DIRECT': ('solver_all_equation_solver', str),

            'RESERVOIR CYCLELENGTH': ('solver_reservoir_cycle_length', float),
            'RESERVOIR MAXCYCLES': ('solver_reservoir_max_cycles', float),
            'RESERVOIR GLOBALTOL': ('solver_reservoir_globaltol', float),
            'RESERVOIR ITERATIVE': ('solver_reservoir_equation_solver', str),
            'RESERVOIR DIRECT': ('solver_reservoir_equation_solver', str),

            'GLOBAL CYCLELENGTH': ('solver_global_cycle_length', float),
            'GLOBAL MAXCYCLES': ('solver_global_max_cycles', float),
            'GLOBAL GLOBALTOL': ('solver_global_globaltol', float),
            'GLOBAL ITERATIVE': ('solver_global_equation_solver', str),
            'GLOBAL DIRECT': ('solver_global_equation_solver', str),

            'NOCUT': ('solver_timestep_cut', bool),
            'CUT': ('solver_timestep_cut', bool),

            'PRECON_ILU': ('solver_precon', str),
            'PRECON_AMG': ('solver_precon', str),
            'PRECON_AMG_RS': ('solver_precon', str),

            'DUAL_SOLVER': ('solver_dual_solver', bool),
            'SYSTEM_REDUCED': ('solver_system_reduced', bool),
            'NBAD': ('solver_nbad', float),
            'PRESSURE_COUPLING': ('solver_pressure_coupling', str),
            'KSUB_METHOD': ('solver_ksub_method', str),
            'FACILITIES': ('solver_facilities', str),
            'PSEUDO_SLACK': ('solver_pseudo_slack', bool),
            'MUMPS_SOLVER': ('solver_mumps_solver', str),
        }
        return solver_keyword_map

    @staticmethod
    def impstab_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the keyword mapping from simulator keyword to ResSimpy attribute and the type of the object.

        Associated with the IMPSTAB keywords in Nexus. e.g. of the form "IMPSTAB TARGETCFL 0.5".
        """
        # IMPSTAB keywords
        impstab_keyword_map = {
            'OFF': ('impstab_on', bool),
            'ON': ('impstab_on', bool),
            'COATS': ('impstab_criteria', str),
            'PEACEMAN': ('impstab_criteria', str),
            'SKIPMASSCFL': ('impstab_skip_mass_cfl', bool),
            'USEMASSCFL': ('impstab_skip_mass_cfl', bool),
            'TARGETCFL': ('impstab_target_cfl', float),
            'LIMITCFL': ('impstab_limit_cfl', float),
            'NOCUTS': ('impstab_no_cuts', float),
            'MAXCUTS': ('impstab_max_cuts', float),
            'SKIPBLOCKDCMAX': ('impstab_skip_block_dcmax', float),
        }
        return impstab_keyword_map

    @staticmethod
    def gridsolver_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the keyword mapping from simulator keyword to ResSimpy attribute and the type of the object.

        Associated with the GRIDSOLVER keywords in Nexus. e.g. of the form "GRIDSOLVER PRESS_RED 0.5".
        """
        # GRIDSOLVER keywords
        gridsolver_keyword_map = {
            'IMPLICIT_COUPLING': ('gridsolver_implicit_coupling_setting', str),
            'IMPES_COUPLING': ('gridsolver_impes_coupling_setting', str),
            'IMPLICIT_PRECON': ('gridsolver_implicit_precon_setting', str),
            'CPR_PRESSURE_EQUATION': ('gridsolver_cpr_pressure_equation', str),
            'PRESS_RED': ('gridsolver_press_reduction', float),
            'IMPLICIT_RED': ('gridsolver_implicit_reduction', float),
            'GRID_RED': ('gridsolver_grid_reduction', float),
        }
        return gridsolver_keyword_map

    @staticmethod
    def solo_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the keyword mapping from simulator keyword to ResSimpy attribute and the type of the object.

        Associated with the keywords that are found on their own in a runcontrol file.".
        """
        # Solo keywords
        solo_keyword_map: dict[str, tuple[str, type]] = {
            'MAXNEWTONS': ('maxnewtons', float),
            'MAXBADNETS': ('maxbadnets', float),
            'CUTFACTOR': ('cutfactor', float),
            'NEGMASSCUT': ('negmasscut', float),
            'DVOLLIM': ('dvollim', float),
            'DZLIM': ('dzlim', float),
            'DSLIM': ('dslim', float),
            'DPLIM': ('dplim', float),
            'DMOBLIM': ('dmoblim', float),
            'DSGLIM': ('dsglim', float),
            'NEGFLOWLIM': ('negflowlim', float),
            'NEGMASSAQU': ('negmassaqu', float),
            'KRDAMP': ('krdamp', float),
            'VOLERR_PREV': ('volerr_prev', float),
            'SGCTOL': ('sgctol', float),
            'EGSGTOL': ('egsgtol', float),
            'SGCPERFTOL': ('sgcperftol', float),
            'LINE_SEARCH': ('line_search', float),
            'PERFP_DAMP': ('perfp_damp', float),
        }
        return solo_keyword_map

    @staticmethod
    def tols_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the keyword mapping from simulator keyword to ResSimpy attribute and the type of the object.

        Associated with the TOLS keywords in Nexus. e.g. of the form "TOLS VOLCON 100.5".
        """
        # TOLS keywords
        tols_keyword_map: dict[str, tuple[str, type]] = {
            'VOLCON': ('tols_volcon', float),
            'MASS': ('tols_mass', float),
            'TARGET': ('tols_target', float),
            'WELLMBAL': ('tols_wellmbal', float),
            'PERF': ('tols_perf', float),
        }
        return tols_keyword_map

    @staticmethod
    def dcmax_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the keyword mapping from simulator keyword to ResSimpy attribute and the type of the object.

        Associated with the DCMAX keywords in Nexus. e.g. of the form "DCMAX IMPES 0.5".
        """
        # DCMAX keywords
        dcmax_keyword_map: dict[str, tuple[str, type]] = {
            'IMPES': ('dcmax_impes', float),
            'IMPLICIT': ('dcmax_implicit', float),
            'ALL': ('dcmax_all', float),
        }
        return dcmax_keyword_map

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the keyword mapping from simulator keyword to ResSimpy attribute and the type of the object.

        Compiled keywords for all ResSimpy attributes apart from the DCMAX keywords.
        """
        # DT keywords
        dt_keyword_map = NexusSolverParameter.dt_keyword_mapping()
        solver_keyword_map = NexusSolverParameter.solver_keyword_mapping()
        gridsolver_keyword_map = NexusSolverParameter.gridsolver_keyword_mapping()
        tols_keyword_map = NexusSolverParameter.tols_keyword_mapping()
        impstab_keyword_map = NexusSolverParameter.impstab_keyword_mapping()
        solo_keyword_map = NexusSolverParameter.solo_keyword_mapping()

        # Method keyword
        misc_keyword_map = {
            'METHOD': ('timestepping_method', TimeSteppingMethod),
            'IMPLICITMBAL': ('implicit_mbal', str),
            'PERFREV': ('perfrev', str),
        }
        # DCMAX keywords
        dcmax_keyword_map = NexusSolverParameter.dcmax_keyword_mapping()

        # Combine the keyword maps
        keyword_map = {**dt_keyword_map, **misc_keyword_map, **solver_keyword_map, **gridsolver_keyword_map,
                       **tols_keyword_map, **impstab_keyword_map, **solo_keyword_map, **dcmax_keyword_map}
        return keyword_map

    @staticmethod
    def get_max_change_attribute_name(keyword: str, value: str) -> tuple[str, type]:
        """Keywords of the form 'DCRPT, DCMAX, D*_MAX_VIP, VOLRPT' have a different attribute name for the value.

        Args:
            keyword (str): The keyword to get the attribute for
            value (str): the value following the keyword to get the attribute for e.g. 'IMPES', 'IMPLICIT', 'ALL'

        Returns:
            str, type: The attribute name for the value, the type of the value
        """
        attribute = keyword.lower()
        return attribute+'_'+value.lower(), float

    @property
    def units(self) -> BaseUnitMapping:
        """Returns the unit system for the NexusSolverParameter which doesn't have units implemented yet."""
        return BaseUnitMapping(unit_system=None)

    def to_string(self) -> str:
        """Outputs the NexusSolverParameters to a string representation for the runcontrol file."""
        param_dict = self.to_dict(keys_in_keyword_style=True, add_date=False, add_units=False, include_nones=False)

        keyword_mapping = self.get_keyword_mapping()

        # handled vars to prevent duplication of entries in the object
        handled_vars = set()
        output_str = ''
        for key, value in param_dict.items():
            variable = keyword_mapping[key][0]
            if value is None or variable in handled_vars:
                continue
            if isinstance(value, bool):
                output_str += self.__handle_truthy_values(variable) + '\n'
            elif key in self.dt_keyword_mapping():
                output_str += f'DT {key} {value}\n'
            elif key in self.solver_keyword_mapping():
                output_str += self.__handle_solver_keywords_to_string(key, value, variable)
            elif key in self.gridsolver_keyword_mapping():
                output_str += f'GRIDSOLVER {key} {value}\n'
            elif key in self.impstab_keyword_mapping():
                output_str += f'IMPSTAB {key} {value}\n'
            elif key in self.tols_keyword_mapping():
                output_str += f'TOLS {key} {value}\n'
            elif key in self.dcmax_keyword_mapping():
                output_str += f'DCMAX {key} {value}\n'
            elif key in self.solo_keyword_mapping():
                output_str += f'{key} {value}\n'
            elif key == 'METHOD' and self.timestepping_method is not None:
                output_str += f'METHOD {self.timestepping_method.value}\n'
            handled_vars.add(variable)
        return output_str + '\n'

    def __handle_solver_keywords_to_string(self, key: str, value: str | float | dict[int, float],
                                           variable: str) -> str:
        """Handles the conversion of solver keywords to string format."""
        if variable == 'solver_all_equation_solver':
            return f'SOLVER ALL {value}\n'
        elif variable == 'solver_reservoir_equation_solver':
            return f'SOLVER RESERVOIR {value}\n'
        elif variable == 'solver_global_equation_solver':
            return f'SOLVER GLOBAL {value}\n'
        elif variable == 'solver_precon':
            return f'SOLVER {value}\n'
        else:
            return f'SOLVER {key} {value}\n'

    def __handle_truthy_values(self, variable: str) -> str:
        """Handles truthy values for NexusSolverParameter."""
        if variable == 'solver_timestep_cut':
            return 'SOLVER CUT' if self.solver_timestep_cut else 'SOLVER NOCUT'
        elif variable == 'solver_dual_solver':
            return 'SOLVER DUAL_SOLVER ON' if self.solver_dual_solver else 'SOLVER DUAL_SOLVER OFF'
        elif variable == 'solver_system_reduced':
            return 'SYSTEM_REDUCED ON' if self.solver_system_reduced else 'SYSTEM_REDUCED OFF'
        elif variable == 'impstab_on':
            return 'IMPSTAB ON' if self.impstab_on else 'IMPSTAB OFF'
        return ''



from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin
from ResSimpy.Enums.FrequencyEnum import FrequencyEnum
from ResSimpy.Enums.OutputType import OutputType
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping


@dataclass(kw_only=True)
class OutputRequest(DataObjectMixin, ABC):
    """Class to hold data input for an Output Request.

    Attributes:
        date (str): Date of the output request.
        output (str): Output request.
        output_type (OutputType): Type of output request.
        output_frequency (FrequencyEnum): Frequency of the output request.
        output_frequency_number (None | float): Number of the frequency of the output request.
    """

    output: str
    output_type: OutputType
    output_frequency: FrequencyEnum
    output_frequency_number: None | float

    def __init__(self, date: str, output: str, output_type: OutputType, output_frequency: FrequencyEnum,
                 output_frequency_number: Optional[float] = None, date_format: Optional[DateFormat] = None,
                 start_date: Optional[str] = None) -> None:
        """Initialises the OutputRequest class.

        Args:
            date (str): Date of the output request.
            output (str): Output request.
            output_type (OutputType): Type of output request.
            output_frequency (FrequencyEnum): Frequency of the output request.
            output_frequency_number (None | float): Number of the frequency of the output request.
            date_format (Optional[DateFormat]): The date format of the object.
            start_date (Optional[str]): The start date of the model (required if the date is in numerical format).
        """

        super().__init__(date=date, date_format=date_format, start_date=start_date)

        self.output = output
        self.output_type = output_type
        self.output_frequency = output_frequency
        self.output_frequency_number = output_frequency_number

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """No keywords for this class, returns an empty dict."""
        return {}

    @property
    def units(self) -> BaseUnitMapping:
        """Writes unit type for the given unit system.

        Returns:
            An instance of the BaseUnitMapping class.
        """
        return BaseUnitMapping(unit_system=None)


@dataclass(kw_only=True)
class OutputContents(DataObjectMixin, ABC):
    """Class to hold data input for an Output Contents.

    Attributes:
        date (str): Date of the output contents.
        output (str): Output contents.
        output_type (OutputType): Type of output contents.
    """

    date: str
    output: str
    output_contents: list[str]
    output_type: OutputType

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """No keywords for this class, returns an empty dict."""
        return {}

    @property
    def units(self) -> BaseUnitMapping:
        """Writes unit type for the given unit system.

        Returns:
            An instance of the BaseUnitMapping class.
        """
        return BaseUnitMapping(unit_system=None)



"""This module contains the abstract base class for file manipulations for simulator files."""
from __future__ import annotations
import os
import pathlib
from datetime import datetime, timezone
from uuid import uuid4, UUID
from dataclasses import dataclass, field
from typing import Optional, Sequence, TypeVar
import warnings
from ResSimpy.FileOperations.FileBase import FileBase
import ResSimpy.FileOperations.file_operations as fo
from ResSimpy.FileOperations.simulator_constants import NEXUS_COMMENT_CHARACTERS, OTHER_SIMULATOR_COMMENT_CHARACTERS
from ResSimpy.Utils.general_utilities import is_number
import uuid

from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_ARRAY_FORMAT_KEYWORDS, GRID_OPERATION_KEYWORDS, \
    GRID_ARRAY_KEYWORDS
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Time.ISODateTime import ISODateTime
from ResSimpy.Utils.factory_methods import get_empty_list_file, get_empty_list_str, get_empty_dict_uuid_list_int

T = TypeVar("T", bound='File')


@dataclass(kw_only=True)
class File(FileBase):
    """The abstract base class for simulator files.

    Attributes:
        location (str): Full path to file location
        file_content_as_list (list[str]): List of lines in the file
        include_locations (Optional[list[str]]): list of file paths that the file contains. Defaults to None.
        origin (Optional[str]): Where the file was opened from. Defaults to None.
        include_objects (Optional[list[NexusFile]]): The include files but generated as a NexusFile instance. \
            Defaults to None.
        linked_user (Optional[str]): user or owner of the file. Defaults to None
        last_modified (Optional[datetime]): last modified date of the file
    """

    location: str
    _location_in_including_file: str
    include_objects: Optional[Sequence[File]]
    include_locations: Optional[list[str]] = None
    file_content_as_list: Optional[list[str]] = field(default=None, repr=False)
    origin: Optional[str] = None
    rootdir: Optional[str] = None
    object_locations: Optional[dict[UUID, list[int]]] = field(default=None, repr=False)
    line_locations: Optional[list[tuple[int, UUID]]] = field(default=None, repr=False)
    linked_user: Optional[str] = field(default=None)
    last_modified: Optional[datetime] = field(default=None)
    __id: UUID = field(default_factory=lambda: uuid4(), compare=False)
    __file_modified: bool = False
    __file_loading_skipped: bool = False

    def __init__(self, location: str,
                 include_locations: Optional[list[str]] = None,
                 origin: Optional[str] = None,
                 rootdir: Optional[str] = None,
                 file_content_as_list: Optional[list[str]] = None,
                 linked_user: Optional[str] = None,
                 last_modified: Optional[datetime] = None,
                 include_objects: Optional[Sequence[File]] = None, create_as_modified: bool = False,
                 file_loading_skipped: bool = False) -> None:
        """Initialises the File class.

        Args:
            location (str): The location of the file on disk.
            file_content_as_list (Optional[list[str]]): The file content as a list of strings.
            include_objects (Optional[Sequence[File]]): The files included in the file.
            create_as_modified (bool): Set the object to modified.
            file_loading_skipped (bool): Whether the file loading was skipped due to the file being too large an array.
            include_locations: Optional[list[str]]: list of file paths to the included files.
            origin: Optional[str]: The file path to the file that included this file. None for top level files.
            rootdir (Optional[str]): The root directory of the main simulation file. Defaults to None.
            include_objects: Optional[list[NexusFile]: list of NexusFile objects that are included in this file.
            file_content_as_list: Optional[list[str]]: list of strings representing the content of the file.
            linked_user (Optional[str]): user or owner of the file. Defaults to None
            last_modified (Optional[datetime]): last modified date of the file, Defaults to None
            file_loading_skipped (bool): If set to True, the file loading was skipped. Defaults to False.
        """
        self._location_in_including_file = location
        self.include_objects: Optional[list[File]] = get_empty_list_file() \
            if include_objects is None else include_objects
        if file_content_as_list is None:
            self.file_content_as_list = []
        else:
            self.file_content_as_list = file_content_as_list

        if origin is not None:
            self.location = fo.get_full_file_path(location, origin, rootdir)
        else:
            self.location = location
        self.include_locations: Optional[list[str]] = get_empty_list_str() if include_locations is None else \
            include_locations
        self.origin: Optional[str] = origin
        self.rootdir: Optional[str] = rootdir
        if self.object_locations is None:
            self.object_locations: dict[UUID, list[int]] = get_empty_dict_uuid_list_int()
        if self.line_locations is None:
            self.line_locations = []
        self.linked_user = linked_user
        self.last_modified = last_modified

        self.__id = uuid.uuid4()
        self.__file_modified = create_as_modified
        self.__file_loading_skipped = file_loading_skipped

    @property
    def id(self) -> UUID:
        """Unique identifier for each Node object."""
        return self.__id

    @property
    def file_modified(self) -> bool:
        """Whether the file has been modified since the last file write."""
        return self.__file_modified

    @property
    def file_loading_skipped(self) -> bool:
        """Whether loading the file contents has been skipped due to it being to large e.g. for array files."""
        return self.__file_loading_skipped

    @property
    def location_in_including_file(self) -> str:
        """The location of the file as it is written after the INCLUDE token in the file including it."""
        return self._location_in_including_file

    def get_content_between_dates(self, start_date: datetime, end_date: datetime,
                                  date_format: DateFormat = DateFormat.MM_DD_YYYY,
                                  model_start_date: Optional[str] = None
                                  ) -> list[str]:
        """Gets all the content in a file between the provided dates as a list of strings.

        Args:
            start_date (datetime): The date to search from.
            end_date (datetime): The date to search to. If not provided, all lines will be returned until the end of the
                                file.
            date_format (DateFormat): The date format of the file. Defaults to the Nexus default (MM/DD/YYYY).
            model_start_date (Optional[str]): The start date of the model. Required if the dates are in decimal format.
        """
        content_between_dates = []

        if self.file_content_as_list is None:
            return []

        time_card_text = 'DATES' if date_format == DateFormat.DD_MMM_YYYY else 'TIME'

        for i, line in enumerate(self.file_content_as_list):
            if fo.check_token(time_card_text, line):
                if time_card_text == 'TIME':
                    date_value = fo.get_expected_token_value(time_card_text, token_line=line,
                                                             file_list=self.file_content_as_list[i:])
                else:
                    date_value = fo.load_in_three_part_date(initial_token='DATES', token_line=line,
                                                            file_as_list=self.file_content_as_list[i:], start_index=0)

                date_value_as_iso = ISODateTime.convert_to_iso(date=date_value, date_format=date_format,
                                                               start_date=model_start_date)

                if date_value_as_iso >= end_date:
                    content_between_dates.append(line)
                    break

                elif date_value_as_iso >= start_date:
                    content_between_dates.append(line)
                    continue

            if any(content_between_dates):
                content_between_dates.append(line)

        return content_between_dates

    def update_include_location_in_file_as_list(self, new_path: str, include_file: File) -> None:
        """update_include_location_in_file_as_list.

        Args:
            new_path(str): Updates the path in the file as list to point towards the new include location.
            include_file(file): include object whose path is being modified.
        """
        raise NotImplementedError("Implement this in the derived class.")

    def write_to_file(self, new_file_path: None | str = None, write_includes: bool = False,
                      write_out_all_files: bool = False, overwrite_file: bool = False) -> None:
        """Writes to file specified in self.location the strings contained in the list self.file_content_as_list.

        Args:
            new_file_path (None | str): writes to self.location if left as None. Otherwise writes to new_file_name.
            write_includes (bool): If True will write out all include files within the file. Defaults to False.
            write_out_all_files (bool): If False will write only modified files. Otherwise will write all files.
            overwrite_file (bool): If True will overwrite the file at the location specified by new_file_path. \
            Otherwise will raise an error if the file already exists.
        """
        # overwrite File base class method to allow for write_includes
        if new_file_path is None:
            if overwrite_file:
                # In this case just overwrite the file with the existing path:
                new_file_path = self.location
            else:
                raise ValueError('No file path to write to, and overwrite_file set to False')

        if self.file_content_as_list is None:
            raise ValueError(f'No file data to write out, instead found {self.file_content_as_list}')

        # create directories that do not exist
        if not os.path.exists(os.path.dirname(new_file_path)):
            os.makedirs(os.path.dirname(new_file_path))

        if write_includes and self.include_objects is not None:
            for file in self.include_objects:
                write_file: bool = file.file_modified or write_out_all_files
                write_file = write_file or (new_file_path != self.location and not os.path.isabs(file.location))

                # if the array was previously skipped then load the file as list
                if file.file_loading_skipped:
                    file.file_content_as_list = fo.load_file_as_list(file.location)

                if file.file_content_as_list is None:
                    warnings.warn(f'No content found for file: {file.location}. Not writing file.')
                    continue
                # this distinguishes the include file name from other includes in other files.
                # this could be moved to a folder name instead rather than affecting the include file name?
                new_root_name = f'{os.path.basename(new_file_path).split(".")[0]}_{os.path.basename(file.location)}'
                # write the include file to the same directory.
                # TODO add a fix for the name above and a check for whether the file exists as exactly the same name
                include_file_name = os.path.join(os.path.dirname(new_file_path), new_root_name)

                if write_file:
                    self.update_include_location_in_file_as_list(include_file_name, file)

                if os.path.exists(include_file_name):
                    # if the file is already copied across then move on to the next file
                    continue

                if write_file:
                    file.write_to_file(include_file_name, write_includes=True, write_out_all_files=write_out_all_files,
                                       overwrite_file=overwrite_file)

        # check if the file already exists so we aren't accidentally overwriting it
        if os.path.exists(new_file_path) and not overwrite_file:
            raise ValueError(f'File already exists at {new_file_path} and overwrite_file set to False')

        file_str = ''.join(self.file_content_as_list)

        # update the location:
        if self.file_modified or write_out_all_files:
            self.location = new_file_path

        # write the file to the new location
        with open(new_file_path, 'w') as fi:
            fi.write(file_str)
        # reset the modified file state
        self._file_modified_set(False)

    @staticmethod
    def get_pathlib_path_details(full_file_path: str) -> None | str:
        """Gets the owner and the group of a particular file.

        Args:
           full_file_path (str): The address of the file.
        """
        if full_file_path == "" or full_file_path is None:
            return None
        pathlib_path = pathlib.Path(full_file_path)
        owner: str = ''
        group: str = ''
        try:
            owner = pathlib_path.owner()  # type: ignore
            group = pathlib_path.group()  # type: ignore
        except NotImplementedError:
            # owner or group not supported on this system, continue without filling out that information
            pass
        except PermissionError:
            # user doesn't have permission to access the file, continue without filling out that information
            warnings.warn(f'PermissionError when trying to access file at {full_file_path}')
        except FileNotFoundError:
            # file not found, continue without filling out that information
            warnings.warn(f'FileNotFoundError when trying to access file at {full_file_path}')
        except KeyError:
            # Group or owner doesn't exist on this system, continue without filling out that information
            warnings.warn(f'Unable to find the group for the file at {full_file_path}')

        if owner is not None and group is not None:
            return f"{owner}:{group}"
        elif owner is not None:
            return owner
        return None

    @staticmethod
    def get_datetime_from_os_stat(full_file_path: str) -> None | datetime:
        """Gets a datetime object representing the last time a file was modified.

        Args:
           full_file_path (str): The address of the file.
        """
        if full_file_path == "" or full_file_path is None:
            return None
        stat_obj = os.stat(full_file_path)
        timestamp = stat_obj.st_mtime
        if timestamp is None:
            return None
        return datetime.fromtimestamp(timestamp, tz=timezone.utc)

    @staticmethod
    def generate_file_include_structure(simulator_type: type[T], file_path: str, origin: Optional[str] = None,
                                        rootdir: Optional[str] = None,
                                        recursive: bool = True, skip_arrays: bool = True,
                                        top_level_file: bool = True) -> T:
        """Generates a nexus file instance for a provided text file with information storing the included files.

        Args:
            simulator_type (type[T]): The type of the Simulator e.g. NexusSimulator
            file_path (str): Path to the file to generate the structure from.
            origin (Optional[str], optional): Where the file was opened from. Defaults to None.
            rootdir (Optional[str], optional): The root directory of the main simulation file. Defaults to None.
            recursive (bool): Whether the method should recursively drill down multiple layers of include_locations.
            skip_arrays (bool): If set True skips the INCLUDE arrays that come after property array and VALUE
            top_level_file (bool): If set to True, the code assumes this is a 'top level' file rather than an included
            one.

        Returns:
            File: a class instance for File with knowledge of include files
        """
        is_nexus_file = simulator_type.__name__ == "NexusFile"  # Can't use isinstance() as importing NexusFile would
        # create a circular reference.
        comment_characters = NEXUS_COMMENT_CHARACTERS if is_nexus_file else OTHER_SIMULATOR_COMMENT_CHARACTERS

        full_file_path = file_path
        if origin is not None:
            full_file_path = fo.get_full_file_path(file_path, origin, rootdir, is_nexus=is_nexus_file)

        if origin is None:
            origin = full_file_path

        try:
            file_as_list = fo.load_file_as_list(full_file_path)
        except FileNotFoundError:
            # handle if a file can't be found
            nexus_file_class = simulator_type(location=file_path,
                                              include_locations=None,
                                              origin=origin,
                                              rootdir=rootdir,
                                              include_objects=None,
                                              file_content_as_list=None,
                                              linked_user=None,
                                              last_modified=None)
            warnings.warn(UserWarning(f'No file found for: {full_file_path} while loading {origin}'))
            return nexus_file_class

        # check last modified and user for the file
        user = File.get_pathlib_path_details(full_file_path)
        last_changed = File.get_datetime_from_os_stat(full_file_path)

        # prevent python from mutating the lists that it's iterating over
        modified_file_as_list: list[str] = []
        # search for the INCLUDE keyword and append to a list:
        inc_file_list: list[str] = []
        includes_objects: Optional[list[T]] = []
        skip_next_include = False

        for i, line in enumerate(file_as_list):
            if is_nexus_file:
                File.__convert_line(full_file_path=full_file_path, is_nexus_file=is_nexus_file, line=line,
                                    modified_file_as_list=modified_file_as_list, simulator_type=simulator_type,
                                    top_level_file=top_level_file)

                should_continue, should_return, skip_next_include = File.__nexus_grid_file_checks(
                    file_as_list=file_as_list, line=line, line_number=i, is_top_level_file=top_level_file,
                    skip_next_include=skip_next_include)

                if should_continue:
                    continue
                elif should_return:
                    nexus_file_class = simulator_type(
                        location=file_path,
                        include_locations=inc_file_list,
                        origin=origin,
                        rootdir=rootdir,
                        include_objects=includes_objects,
                        file_content_as_list=modified_file_as_list
                    )

                    return nexus_file_class

            else:
                modified_file_as_list.append(line)

            if not fo.check_token(token="INCLUDE", line=line, comment_characters=comment_characters):
                # No include on this line, go to the next one.
                continue

            inc_file_path = fo.get_token_value(token='INCLUDE', token_line=line, file_list=file_as_list[i:],
                                               comment_characters=comment_characters, single_c_comments=is_nexus_file,
                                               remove_quotation_marks=True)
            if inc_file_path is None:
                continue
            inc_full_path = fo.get_full_file_path(inc_file_path, origin=full_file_path, rootdir=rootdir,
                                                  is_nexus=is_nexus_file)
            # store the included files as files inside the object
            inc_file_list.append(inc_full_path)

            # test the include to see if the first few lines have only array data
            # limit number of lines loaded here in future?
            if skip_arrays:
                try:
                    inc_file_as_list = fo.load_file_as_list(inc_full_path)
                except FileNotFoundError:
                    # handle files not found - this is handled in an exception in the main loop
                    pass
                else:
                    all_numeric = False
                    for inc_file_line in inc_file_as_list[0:50]:
                        split_line = fo.split_line(inc_file_line, upper=False)
                        # check if it is numeric data
                        # this won't work if the array has scientific notation.
                        if any(not is_number(x) for x in split_line):
                            # don't set skip_next_include if the line is not entirely numeric
                            all_numeric = False
                            break
                        all_numeric = True
                    if all_numeric:
                        skip_next_include = True

            if not recursive:
                continue
            elif skip_arrays and skip_next_include:
                inc_file = simulator_type(location=inc_file_path,
                                          include_locations=None,
                                          origin=full_file_path,
                                          rootdir=rootdir,
                                          include_objects=None,
                                          file_content_as_list=None,
                                          linked_user=user,
                                          last_modified=last_changed,
                                          file_loading_skipped=True)
                if includes_objects is None:
                    raise ValueError('include_objects is None - recursion failure.')
                skip_next_include = False
            else:
                inc_file = simulator_type.generate_file_include_structure(simulator_type=simulator_type,
                                                                          file_path=inc_file_path,
                                                                          origin=full_file_path, rootdir=rootdir,
                                                                          recursive=True,
                                                                          skip_arrays=skip_arrays, top_level_file=False)
                if includes_objects is None:
                    raise ValueError('include_objects is None - recursion failure.')

            includes_objects.append(inc_file)

        includes_objects = None if not includes_objects else includes_objects

        nexus_file_class = simulator_type(
            location=file_path,
            include_locations=inc_file_list,
            origin=origin,
            rootdir=rootdir,
            include_objects=includes_objects,
            file_content_as_list=modified_file_as_list,
            linked_user=user,
            last_modified=last_changed
        )

        return nexus_file_class

    @staticmethod
    def __nexus_grid_file_checks(line: str, file_as_list: list[str], line_number: int, is_top_level_file: bool,
                                 skip_next_include: bool) -> tuple[bool, bool, bool]:
        """Checks a file to see if it is a grid array file, that doesn't need to be loaded in."""
        if line.rstrip().endswith('>'):
            return True, False, skip_next_include

        if fo.check_token("INCLUDE", line):
            # Include found, check if we should skip loading it in (e.g. if it is a large array file)
            ignore_keywords = ['NOLIST']
            previous_value = fo.get_previous_value(file_as_list=file_as_list[0: line_number + 1],
                                                   search_before='INCLUDE', ignore_values=ignore_keywords)

            keywords_to_skip_include = GRID_ARRAY_FORMAT_KEYWORDS + GRID_OPERATION_KEYWORDS + ["CORP"]
            if previous_value is None:
                skip_next_include = False

            elif previous_value.upper() in keywords_to_skip_include:
                skip_next_include = True

            return False, False, skip_next_include

        elif fo.check_token("VALUE", line) and not is_top_level_file:
            # Check if this is an 'embedded' grid array file. If it is, return this file with only the content up
            # to this point to help with performance when analysing the files.
            previous_value = fo.get_previous_value(file_as_list=file_as_list[0: line_number + 1], search_before='VALUE')
            next_value = fo.get_next_value(start_line_index=0, file_as_list=file_as_list[line_number:],
                                           search_string=line.upper().split('VALUE')[1])

            if previous_value is None or next_value is None:
                return True, False, skip_next_include

            if next_value.upper() != 'INCLUDE' and previous_value.upper() in GRID_ARRAY_KEYWORDS:
                return False, True, skip_next_include
            else:
                return True, False, skip_next_include

        else:
            return True, False, skip_next_include

    @staticmethod
    def __convert_line(full_file_path: str, is_nexus_file: bool, line: str, modified_file_as_list: list[str],
                       simulator_type: type[T], top_level_file: bool) -> None:
        """Modifies a line in a text file to allow easier loading of the contents."""
        if len(modified_file_as_list) >= 1:
            previous_line = modified_file_as_list[len(modified_file_as_list) - 1].rstrip()
            # Handle lines continued with the '>' character
            if is_nexus_file and previous_line.endswith('>'):
                modified_file_as_list[len(modified_file_as_list) - 1] = previous_line[:-1] + line
            else:
                if not top_level_file:
                    converted_line = simulator_type.convert_line_to_full_file_path(line=line,
                                                                                   full_base_file_path=full_file_path)
                else:
                    converted_line = line
                modified_file_as_list.append(converted_line)
        else:
            if not top_level_file:
                converted_line = simulator_type.convert_line_to_full_file_path(line=line,
                                                                               full_base_file_path=full_file_path)
            else:
                converted_line = line
            modified_file_as_list.append(converted_line)

    def _file_modified_set(self, value: bool) -> None:
        """Set the modified file status.

        Args:
            value(bool): True if the file is modified, False otherwise.
        """
        self.__file_modified = value

    @property
    def get_flat_list_str_file(self) -> list[str]:
        """Returns flat list of strings from file."""
        raise NotImplementedError("Implement this in the derived class")

    def add_object_locations(self, obj_uuid: UUID, line_indices: list[int]) -> None:
        """Adds a uuid to the object_locations dictionary.

        Args:
            obj_uuid(UUID): Unique identifier of the object being created/stored.
            line_indices(list[int]): line number in the flattened file_content_as_list
            (i.e. from the get_flat_list_str_file method).
        """
        raise NotImplementedError("Implement this in the derived class")

    @staticmethod
    def insert_comments(additional_content: list[str], comments: str) -> list[str]:
        """Adds comments alongside additional content.

        Args:
            additional_content(list[str]): additional lines of the file to be added with a new entry per line.
            comments(str): comments to be added to all lines list of strings within the content.
        """
        raise NotImplementedError("Implement this in the derived class")

    def get_object_locations_for_id(self, obj_id: UUID) -> list[int]:
        """Returns the list of line numbers associated with the object ID.

        Args:
            obj_id(UUID): Unique identifier for each node object.
        """
        raise NotImplementedError("Implement this in the derived class")

    def remove_object_from_file_as_list(self, objects_to_remove: list[UUID]) -> None:
        """Removes all associated lines in the file as well as the object locations relating to a list of objects.

        Args:
            objects_to_remove(list[UUID]):  list of object ids to remove from the object locations. Defaults to None
        """
        raise NotImplementedError("Implement this in the derived class")

    def add_to_file_as_list(self, additional_content: list[str], index: int,
                            additional_objects: Optional[dict[UUID, list[int]]] = None,
                            comments: Optional[str] = None) -> None:
        """To add content to the file as list, also updates object numbers
        and optionally allows user to add several additional new objects.

        Args:
            additional_content(list[str]):  Additional lines as a list of strings to be added.
            index(int): index to insert the new lines at in the calling flat_file_as_list
            additional_objects(Optional[dict[UUID, list[int]]]): defaults to None. Otherwise,
            a dictionary keyed with the UUID of the new objects to add as well as
            the corresponding index of the object in the original calling NexusFile.
            comments(Optional[str]): defaults to None. Comments to add in-line to the file.
        """
        raise NotImplementedError("Implement this in the derived class")

    def remove_from_file_as_list(self, index: int, objects_to_remove: Optional[list[UUID]] = None,
                                 string_to_remove: Optional[str] = None) -> None:
        """Remove an entry from the file as list.

        Args:
            index(int): index n the calling flat_file_as_list to remove the entry from.
            objects_to_remove(Optional[list[UUID]): list of object ids to
            remove from the object locations. Defaults to None.
            string_to_remove(Optional[str]):if specified will only remove the listed string from the entry at the index.
             Defaults to None, which removes the entire entry.
        """
        raise NotImplementedError("Implement this in the derived class")

    def find_which_include_file(self, flattened_index: int) -> tuple[File, int]:
        """Given a line index that relates to a position within the flattened
        file_as_list from the method get_flat_file_as_list.

        Args:
            flattened_index(int): index in the flattened file as list structure.
        """
        raise NotImplementedError("Implement this in the derived class")

    def __repr__(self) -> str:
        """A more readable representation of the NexusFile object file data."""
        repr_string = ''

        repr_string += f'FILE PATH: {self.location}\n\n'
        repr_string += f'Include files: {self.include_locations}\n\n'

        repr_string += 'FILE CONTENTS:\n\n'
        repr_string += self.pretty_print_contents()
        return repr_string

    def pretty_print_contents(self) -> str:
        """Pretty print the file contents."""
        if self.file_content_as_list is None:
            return ''
        return ''.join(self.file_content_as_list)

    def get_include_file_from_filename(self, filename: str) -> list[File]:
        """Retrieves a list of files matching the provided file name.

        Args:
            filename(str): file name to search for.
        """
        matching_files: list[File] = []

        if filename in self.location:
            matching_files.append(self)

        if self.include_objects is None:
            return matching_files

        for file in self.include_objects:
            if filename in file.location:
                matching_files.append(file)
            elif file.include_objects is not None and any(file.include_objects):
                # If there are includes within this included file, search them as well.
                embedded_matching_files = file.get_include_file_from_filename(filename=filename)
                if any(embedded_matching_files):
                    matching_files.extend(embedded_matching_files)

        return matching_files

    @staticmethod
    def convert_line_to_full_file_path(line: str, full_base_file_path: str) -> str:
        """Modifies a file reference to contain the full file path for easier loading later."""
        raise NotImplementedError("Implement in the inheriting class")



from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.Over import Over


@dataclass(kw_only=True)
class NexusTOver(Over):
    """Used to represent an Override for transmissibility arrays using values per cell.

    Attributes:
        array (str): The array to be modified by the over.
        i1 (int): The start of the i range.
        i2 (int): The end of the i range.
        j1 (int): The start of the j range.
        j2 (int): The end of the j range.
        k1 (int): The start of the k range.
        k2 (int): The end of the k range.
        operator (str): The operator to be assigned to the range.
        grid (Optional[str]): The grid that the Over function relates to.
        include_file (str): The include file that the TOver function reads values from.
    """
    array: str
    i1: int
    i2: int
    j1: int
    j2: int
    k1: int
    k2: int
    operator: str
    include_file: Optional[str]
    grid: Optional[str] = None
    array_values: Optional[list[float]] = None



from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat


@dataclass(kw_only=True)
class NetworkList(DataObjectMixin, ABC):
    """Abstract Base Class for representing a single NetworkList or group for the model."""
    elements_in_the_list: list[str]

    def __init__(self, date: str, elements_in_the_list: list[str], name: str, date_format: Optional[DateFormat] = None,
                 start_date: Optional[str] = None) -> None:
        """Initialises the NetworkList class.

        Args:
            date (str): Date that the con list is created.
            elements_in_the_list (list[str]): A list of the names of the objects in the list.
            name ([str]): The name of the NetworkList.
            date_format (Optional[DateFormat]): The date format of the object.
            start_date (Optional[str]): The start date of the model (required if the date is in numerical format).
        """

        super().__init__(date=date, date_format=date_format, start_date=start_date, name=name)

        self.elements_in_the_list = elements_in_the_list



from __future__ import annotations
from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from typing import TYPE_CHECKING, TypeAlias

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusAquiferMethods import NexusAquiferMethods
    from ResSimpy.Nexus.NexusEquilMethods import NexusEquilMethods
    from ResSimpy.Nexus.NexusGasliftMethods import NexusGasliftMethods
    from ResSimpy.Nexus.NexusHydraulicsMethods import NexusHydraulicsMethods
    from ResSimpy.Nexus.NexusPVTMethods import NexusPVTMethods
    from ResSimpy.Nexus.NexusRelPermMethods import NexusRelPermMethods
    from ResSimpy.Nexus.NexusRockMethods import NexusRockMethods
    from ResSimpy.Nexus.NexusSeparatorMethods import NexusSeparatorMethods
    from ResSimpy.Nexus.NexusValveMethods import NexusValveMethods
    from ResSimpy.Nexus.NexusWaterMethods import NexusWaterMethods

    DynamicMethodCollection: TypeAlias = (
            NexusRockMethods | NexusPVTMethods | NexusEquilMethods | NexusWaterMethods |
            NexusSeparatorMethods | NexusHydraulicsMethods | NexusAquiferMethods | NexusValveMethods |
            NexusGasliftMethods | NexusRelPermMethods
               )


def add_dynamic_method(dynamic_method_collection: DynamicMethodCollection, method: DynamicProperty, new_file_name: str,
                       create_new_file: bool = False) -> None:
    """Adds a new DynamicMethod to the collection.

    Args:
        dynamic_method_collection (DynamicMethodCollection): The collection to add the method to.
        method (DynamicProperty): The dynamic method to add.
        new_file_name (str): The name of the file to save the method to.
        create_new_file (bool): Whether to create a new file for the method.
    """
    if new_file_name is None or new_file_name.strip() == '':
        raise ValueError('New file name must be provided and cannot be empty when adding a new method.')

    if method.input_number in dynamic_method_collection.inputs:
        raise ValueError(f'Method with input number {method.input_number} already exists in the collection.')

    method_type = dynamic_method_collection._method_type()

    if not isinstance(method, method_type):
        raise TypeError(f'Expected method of type {method_type.__name__}, but got {type(method).__name__}.')

    dynamic_method_collection.__properties_loaded = True  # We are adding a new method so no more loading can happen.

    new_nexus_file = NexusFile(location=new_file_name,
                               origin=None,  # Should get connected to the parent fcs file later
                               include_objects=None,
                               file_content_as_list=method.to_string().splitlines(keepends=True)
                               )
    new_nexus_file._file_modified_set(True)  # Mark the file as modified
    method.file = new_nexus_file

    if create_new_file:
        method.write_to_file(new_file_path=new_file_name, overwrite_file=False)

    # explicit guard against the method type is used above so ignore mypy.
    dynamic_method_collection.inputs[method.input_number] = method  # type: ignore[assignment]
    dynamic_method_collection.files[method.input_number] = method.file
    if dynamic_method_collection._model_files is not None:
        dynamic_method_collection._model_files._add_file(new_nexus_file, keyword=dynamic_method_collection.keyword,
                                                         method_number=method.input_number)



"""Class for collection of Nexus water property methods."""
from __future__ import annotations
from dataclasses import dataclass, field
import os
from typing import Optional, MutableMapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusWaterMethod import NexusWaterMethod
from ResSimpy.DataModelBaseClasses.Water import Water
from ResSimpy.Utils.dynamic_method_manipulations import add_dynamic_method


@dataclass(kw_only=True)
class NexusWaterMethods(Water):
    """Class for collection of Nexus water property methods.

    Attributes:
        inputs (dict[int, NexusWaterMethod]): Collection of Nexus water property methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of water property files, as defined in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusWaterMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem
    _model_files: Optional[FcsNexusFile] = field(default=None, repr=False, compare=False)

    def __init__(self, model_unit_system: UnitSystem,
                 inputs: Optional[MutableMapping[int, NexusWaterMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None, assume_loaded: bool = False,
                 model_files: Optional[FcsNexusFile] = None) -> None:
        """Initialises the NexusWaterMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusWaterMethod]]): Collection of Nexus water property methods.
            files (Optional[dict[int, NexusFile]]): Collection of water property files, as defined in Nexus fcs file.
                Keyed by the method number.
            assume_loaded (bool): If True, assumes that the water methods have already been loaded.
            model_files (Optional[FcsNexusFile]): The FCS file containing the whole set of Nexus files.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusWaterMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        self.__properties_loaded = assume_loaded
        self._model_files = model_files
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing water methods."""
        if not self.__properties_loaded:
            self.load_water_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'WATER method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusWaterMethod]:
        """Returns mapping of water property methods as a dictionary where keys are the method number of type int."""
        if not self.__properties_loaded:
            self.load_water_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        """Returns dictionary of 'NexusFile' objects where keys are of type int."""
        return self.__files

    def load_water_methods(self) -> None:
        """Loads water properties from files and initialises NexusWaterMethod object."""
        # Read in water properties from Nexus water method files
        if self.__files is not None and len(self.__files) > 0:  # Check if water files exist
            for table_num in self.__files.keys():  # For each water property method
                water_file = self.__files[table_num]
                if water_file.location is None:
                    raise ValueError(f'Unable to find water file: {water_file}')
                if os.path.isfile(water_file.location):
                    # Create NexusWaterMethod object
                    self.__inputs[table_num] = NexusWaterMethod(file=water_file, input_number=table_num,
                                                                model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with water properties in file
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system

    @property
    def keyword(self) -> str:
        """Return the keyword for the water methods."""
        return 'WATER'

    def add_method(self, method: DynamicProperty, new_file_name: str, create_new_file: bool = False) -> None:
        """Adds a new Water method to the collection.

        Args:
            method (DynamicProperty): The water method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        add_dynamic_method(dynamic_method_collection=self, method=method, new_file_name=new_file_name,
                           create_new_file=create_new_file)

    def _method_type(self) -> type[NexusWaterMethod]:
        """Returns the expected type of the dynamic property."""
        return NexusWaterMethod



"""Holds a class for representing a single ConList or group for the Nexus model."""
from dataclasses import dataclass

from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
from ResSimpy.DataModelBaseClasses.NetworkList import NetworkList


@dataclass(kw_only=True, init=False)
class NexusConList(NetworkList):
    """Class for representing a single ConList or group for the Nexus model."""

    def __init__(self, name: str, elements_in_the_list: list[str], date: str, date_format: DateFormat) -> None:
        """Initialises the NexusConList class.

        Args:
            name (str): Name of the conlist.
            elements_in_the_list (list[str]): List of connection names in the conlist.
            date (str): Date when the conlist is defined. Persists until the next date is defined.
            date_format (Optional[DateFormat]): The date format of the object.
        """
        super().__init__(name=name, elements_in_the_list=elements_in_the_list, date=date, date_format=date_format)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Returns mapping of keywords."""
        return {}

    @property
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the data object."""
        return BaseUnitMapping(None)

    @property
    def connections(self) -> list[str]:
        """Returns the list of connections in the conlist."""
        return self.elements_in_the_list



"""Data object storing the solver parameters at a given datetime for the Simulation runs."""

from dataclasses import dataclass
from abc import ABC
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin


@dataclass
class SolverParameter(DataObjectMixin, ABC):
    """Abstract Data object storing the solver parameters at a given datetime for the Simulation runs.

    Stored as a datetime dependent object as some sims can change these parameters with time.

    Attributes:
        date (str): The date at which these parameters are valid.
        dsrdt_limit (Optional[float]): DRSDT limit to be set.
        dsrdt_two_phases (Optional[bool]): If True, limit is only applied to cells where both oil and gas phases exist.
    """
    date: str
    drsdt_limit: Optional[float] = None
    drsdt_two_phases: Optional[bool] = None



from __future__ import annotations

import warnings
from typing import Optional
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from datetime import datetime, timedelta


class ISODateTime(datetime):
    """A class representing an extension of datetime class,  returns back date in ISO datetime format."""

    def __repr__(self) -> str:
        """Return the object representation, but formatted in ISO format."""
        basic_string = super().__repr__()
        iso_string = basic_string.replace(' ', 'T')
        return iso_string

    def __str__(self) -> str:
        """Return the string representation, but formatted in ISO format."""
        basic_string = super().__str__()
        iso_string = basic_string.replace(' ', 'T')
        return iso_string

    @staticmethod
    def isfloat(no_of_days: str) -> bool:
        """Checks if provided string can be turned into a float.

        Args:
            no_of_days(str): Number of days.
        """
        if no_of_days is not None:
            try:
                float(no_of_days)
                return True
            except ValueError:
                return False
        else:
            return False

    @classmethod
    def convert_to_iso(cls: type[ISODateTime], date: str, date_format: Optional[DateFormat],
                       start_date: Optional[str] = None) -> ISODateTime:
        """Converts an ordinary date to an ISODate format.

        Args:
            date (str): The date as it is written in the model file.
            date_format (DateFormat): The date format to use to convert the date.
            start_date (Optional[str]): The start date of the model (required if the date is a number of days from the
                start).

        Raises:
            ValueError: if the provided parameters cannot produce a valid date.

        Returns:
        ISODateTime: an initialised ISODateTime class instance for the date provided.
        """
        converted_date = None

        if date_format is None:
            warnings.warn(f"No date format provided for date at {date}, assuming MM/DD/YYYY")
            date_format = DateFormat.MM_DD_YYYY

        if ISODateTime.isfloat(date) and start_date is None:
            raise ValueError(f'Found date: "{date}". Please provide start date when date is numeric')
        elif ISODateTime.isfloat(date) and start_date is not None:
            if date_format == DateFormat.DD_MM_YYYY:
                converted_date = ISODateTime.strptime(start_date, '%d/%m/%Y') + timedelta(days=float(date))
            elif date_format == DateFormat.MM_DD_YYYY:
                converted_date = ISODateTime.strptime(start_date, '%m/%d/%Y') + timedelta(days=float(date))

        elif date_format == DateFormat.DD_MM_YYYY:
            try:
                converted_date = ISODateTime.strptime(date, '%d/%m/%Y')
            except ValueError:  # Handling the case where a time has been added
                converted_date = ISODateTime.strptime(date, '%d/%m/%Y(%H:%M:%S)')

        elif date_format == DateFormat.MM_DD_YYYY:
            try:
                converted_date = ISODateTime.strptime(date, '%m/%d/%Y')
            except ValueError:  # Handling the case where a time has been added
                converted_date = ISODateTime.strptime(date, '%m/%d/%Y(%H:%M:%S)')

        elif date_format == DateFormat.DD_MMM_YYYY:
            try:
                converted_date = ISODateTime.strptime(date, '%d %b %Y')
            except ValueError:  # Handling the case where a time has been added
                if len(date) < 21:
                    converted_date = ISODateTime.strptime(date, '%d %b %Y %H:%M:%S')
                else:
                    converted_date = ISODateTime.strptime(date, '%d %b %Y %H:%M:%S.%f')

        elif date_format == DateFormat.DD_MM_YYYY_h_m_s:
            converted_date = ISODateTime.strptime(date, '%d/%m/%Y(%H:%M:%S)')

        elif date_format == DateFormat.MM_DD_YYYY_h_m_s:
            converted_date = ISODateTime.strptime(date, '%m/%d/%Y(%H:%M:%S)')

        if converted_date is None:
            raise ValueError('Invalid date format or missing start_date.')

        return converted_date

    @classmethod
    def datetime_to_iso(cls: type[ISODateTime], date: datetime, datetime_format: str = '%Y-%m-%d') -> ISODateTime:
        """Converts datetime object to ISODateTime object."""
        return ISODateTime.strptime(str(date), datetime_format)

    def strftime_dateformat(self, new_date_format: DateFormat) -> str:
        """Converts a date string to the requested format.

        Args:
        new_date_format (DateFormat): The date format to convert to.

        Returns:
        str: The converted date
        """

        # check if the iso time is non-zero
        if self.time() != datetime.min.time():
            time_included = True
        else:
            time_included = False

        match new_date_format:
            case DateFormat.DD_MMM_YYYY:
                if time_included:
                    new_date_str = datetime.strftime(self, '%d %b %Y %H:%M:%S').lstrip('0').upper()
                else:
                    new_date_str = datetime.strftime(self, '%d %b %Y').lstrip('0').upper()
            case DateFormat.DD_MM_YYYY:
                if time_included:
                    new_date_str = datetime.strftime(self, '%d/%m/%Y(%H:%M:%S)')
                else:
                    new_date_str = datetime.strftime(self, '%d/%m/%Y')
            case DateFormat.MM_DD_YYYY:
                if time_included:
                    new_date_str = datetime.strftime(self, '%m/%d/%Y(%H:%M:%S)')
                else:
                    new_date_str = datetime.strftime(self, '%m/%d/%Y')
            case _:
                raise NotImplementedError("Requested conversion not implemented yet.")

        return new_date_str



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class Wellbore(NetworkObject, ABC):
    diameter: Optional[float] = None
    inner_diameter: Optional[float] = None
    roughness: Optional[float] = None

    def __init__(self, properties_dict: Optional[DataObjectMixinDictType] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None, diameter: Optional[float] = None,
                 inner_diameter: Optional[float] = None, roughness: Optional[float] = None) -> None:
        """Initialises the Wellbore class.

        Args:
            properties_dict (dict): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the object.
            diameter (Optional[float]): Diameter of the well.(DIAM)
            inner_diameter (Optional[float]): Inner diameter of the well. (INNERDIAM)
            roughness (Optional[float]): Roughness of the well. (ROUGHNESS)
        """

        self.diameter = diameter
        self.inner_diameter = inner_diameter
        self.roughness = roughness

        super().__init__(properties_dict=properties_dict, date=date, date_format=date_format, start_date=start_date,
                         unit_system=unit_system, name=name)

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the WellConnection."""
        return NetworkUnits(self.unit_system)



"""Class to represent grid array functions."""
from dataclasses import dataclass
from ResSimpy.Enums.GridFunctionTypes import GridFunctionTypeEnum
from typing import Optional


@dataclass(kw_only=True)
class GridArrayFunction:
    """Class to represent grid array functions.

    Attributes:
        region_type (str): Type of regions in which function input option is applied
        region_number (list[int]): Region numbers, to which function input options are to be applied
        input_array (list[str]): Name of input array(s)
        output_array (list[str]): Name of output array(s)
        function_type (GridFunctionTypeEnum): Analytical function type for function that acts on input arrays
        function_values (list[float]): Coefficients of polynomial or step function
        blocks (list[int]): Range of grid blocks, [i1, i2, j1, j2, k1, k2]
        grid_name (str): Name of grid to which function applies
        input_range (list[tuple[float, float]]): Min-max input range tuples of the input variables
        output_range (list[tuple[float, float]]): Min-max output range tuples of the output variables
        drange (list[float]): List of maximum distances from the input point
    """
    region_type: Optional[str] = None
    region_number: Optional[list[int]] = None
    input_array: Optional[list[str]] = None
    output_array: Optional[list[str]] = None
    function_type: Optional[GridFunctionTypeEnum] = None
    function_values: Optional[list[float]] = None
    blocks: Optional[list[int]] = None
    grid_name: Optional[str] = None
    input_range: Optional[list[tuple[float, float]]] = None
    output_range: Optional[list[tuple[float, float]]] = None
    drange: Optional[list[float]] = None



"""Holds the class for the NexusSolverParams which handles loading and storing of the series of \
NexusSolverParam data objects.
"""
from __future__ import annotations

from typing import Sequence, TYPE_CHECKING

from ResSimpy.Enums.TimeSteppingMethodEnum import TimeSteppingMethod
from ResSimpy.Nexus.DataModels.NexusSolverParameter import NexusSolverParameter
from ResSimpy.Nexus.NexusKeywords.runcontrol_keywords import (DT_KEYWORDS, SOLVER_SCOPE_KEYWORDS,
                                                              SOLVER_SCOPED_KEYWORDS, SOLVER_KEYWORDS,
                                                              IMPSTAB_KEYWORDS, GRIDSOLVER_KEYWORDS, SOLO_KEYWORDS,
                                                              TOLS_KEYWORDS, DCMAX_KEYWORDS, MAX_CHANGE_KEYWORDS)
from ResSimpy.GenericContainerClasses.SolverParameters import SolverParameters
from ResSimpy.FileOperations import file_operations as fo
from ResSimpy.Time.ISODateTime import ISODateTime

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


class NexusSolverParameters(SolverParameters):
    def __init__(self, model: NexusSimulator, assume_loaded: bool = False) -> None:
        """NexusSolverParameters class constructor.

        Args:
            model (NexusSimulator): Originating NexusSimulator object.
            assume_loaded (bool): If True, assumes that the solver parameters are already loaded and does not attempt
                to load them.
        """
        self.__solver_parameters: Sequence[NexusSolverParameter] | None = None
        self.__model = model
        self.file_content = ['']
        self.start_date = ''
        self.__assume_loaded = assume_loaded

    @property
    def solver_parameters(self) -> Sequence[NexusSolverParameter]:
        """Returns a list of solver parameters (usually of type list).

        If solver parameters are not loaded 'self.load' will attempt to load them.
        If loading fails it will return a value error.
        """
        if self.__solver_parameters is None and not self.__assume_loaded:
            self.load()
        if self.__solver_parameters is None:
            return []
        return self.__solver_parameters

    def get_all(self) -> Sequence[NexusSolverParameter]:
        """Returns all solver parameters."""
        return self.solver_parameters

    def set_solver_parameters(self, value: list[NexusSolverParameter]) -> None:
        """Sets solver parameters for this instance.

        Args:
            value(list[NexusSolverParameter]): list of solver parameters.
        """
        self.__solver_parameters = value

    def load(self) -> None:
        """Loads data from run control file and sets start date from the model."""
        if self.__model.model_files.runcontrol_file is None:
            raise ValueError('No runcontrol file found when trying to load solver parameters.')
        self.file_content = self.__model.model_files.runcontrol_file.get_flat_list_str_file
        self.start_date = self.__model.start_date

        read_in_solver_parameter: list[NexusSolverParameter] = []
        # read in the solver parameters from the runcontrol file

        current_solver_param_token = None
        current_date = self.start_date
        solver_parameter_for_timestep = NexusSolverParameter(date=self.start_date)
        current_solver_scope = 'ALL'
        solver_parameters_that_work_with_generic_function = {
            'DT': DT_KEYWORDS,
            'GRIDSOLVER': GRIDSOLVER_KEYWORDS,
            'TOLS': TOLS_KEYWORDS,
        }
        solver_parameters_that_work_with_generic_function.update({key: DCMAX_KEYWORDS for key in MAX_CHANGE_KEYWORDS})

        for line in self.file_content:
            # maybe put a for loop in here for all the potential starting tokens?
            if fo.check_token(token='TIME', line=line):
                # append the existing solver parameter to the list
                if solver_parameter_for_timestep != NexusSolverParameter(date=current_date):
                    # prevent blank solver parameters from being added to the list
                    read_in_solver_parameter.append(solver_parameter_for_timestep)
                # create a new solver parameter object for the new time block
                current_date = fo.get_expected_token_value('TIME', line, file_list=self.file_content)
                solver_parameter_for_timestep = NexusSolverParameter(date=current_date)
                # reset the current_solver_param_token
                current_solver_param_token = None

            # see if we get any SOLVER blocks
            if fo.check_token(token='SOLVER', line=line):
                current_solver_param_token = 'SOLVER'
                solver_token_value = fo.get_expected_token_value('SOLVER', line, file_list=self.file_content)
                solver_parameter_for_timestep = (
                    self.__get_solver_token_values(solver_token_value, line, solver_parameter_for_timestep))
                current_solver_scope = solver_token_value

            if current_solver_param_token == 'SOLVER' and not fo.check_token(token='SOLVER', line=line):
                solver_parameter_for_timestep = self.__set_solver_parameters(current_solver_scope, line,
                                                                             solver_parameter_for_timestep)

            if fo.check_token(token='METHOD', line=line):
                timestep_method = fo.get_expected_token_value('METHOD', line, file_list=self.file_content)
                # convert the string to the enum
                if timestep_method.lower() == 'implicit':
                    timestep_method_enum = TimeSteppingMethod.IMPLICIT
                else:
                    timestep_method_enum = TimeSteppingMethod.IMPES
                # set the timestepping_method in the object
                solver_parameter_for_timestep.timestepping_method = timestep_method_enum

            if fo.check_token(token='IMPLICITMBAL', line=line):
                solver_parameter_for_timestep.implicit_mbal = fo.get_expected_token_value('IMPLICITMBAL', line,
                                                                                          file_list=self.file_content)

            if fo.check_token(token='IMPSTAB', line=line):
                current_solver_param_token = 'IMPSTAB'
                impstab_token_value = fo.get_token_value('IMPSTAB', line, file_list=self.file_content)
                if impstab_token_value is not None:
                    solver_parameter_for_timestep.impstab_on = impstab_token_value.upper() == 'ON'

            if current_solver_param_token == 'IMPSTAB' and not fo.check_token(token='IMPSTAB', line=line):
                next_value = fo.get_next_value(0, file_as_list=[line])
                if next_value is not None and next_value in IMPSTAB_KEYWORDS:
                    solver_parameter_for_timestep = self.__get_impstab_token_values(next_value, line,
                                                                                    solver_parameter_for_timestep)

            for possible_solver_param_tokens in solver_parameters_that_work_with_generic_function:
                if fo.check_token(token=possible_solver_param_tokens, line=line):
                    current_solver_param_token = possible_solver_param_tokens
                    grid_solver_method = fo.get_expected_token_value(current_solver_param_token, line, file_list=[line])
                    solver_parameter_for_timestep = (
                        self.__get_generic_solver_token_values(grid_solver_method, line, solver_parameter_for_timestep,
                                                               current_solver_param_token))

            if (current_solver_param_token in solver_parameters_that_work_with_generic_function and
                    not fo.check_token(token=current_solver_param_token, line=line)):
                next_value = fo.get_next_value(0, file_as_list=[line])
                valid_keywords = solver_parameters_that_work_with_generic_function[current_solver_param_token]
                if next_value is not None and next_value.upper() in valid_keywords:
                    solver_parameter_for_timestep = (
                        self.__get_generic_solver_token_values(next_value, line, solver_parameter_for_timestep,
                                                               current_solver_param_token))
            if fo.check_token(token='PERFREV', line=line):
                solver_parameter_for_timestep.perfrev = fo.get_expected_token_value('PERFREV', line,
                                                                                    file_list=self.file_content)

            if fo.check_token(token='DRSDT', line=line):
                drsdt_limit = fo.get_expected_token_value('LIMIT', line,
                                                          file_list=self.file_content)
                solver_parameter_for_timestep.drsdt_limit = float(drsdt_limit)
                if "2PHASE" in line:
                    solver_parameter_for_timestep.drsdt_two_phases = True

            for keyword in SOLO_KEYWORDS:
                if fo.check_token(token=keyword, line=line):
                    self.__get_generic_solver_token_values(keyword, line, solver_parameter_for_timestep,
                                                           'SOLO')

        read_in_solver_parameter.append(solver_parameter_for_timestep)

        # finally assign the read in solver parameters to the class variable
        self.__solver_parameters = read_in_solver_parameter

    def __set_solver_parameters(self, current_solver_scope: str, line: str,
                                solver_parameter_for_timestep: NexusSolverParameter) -> NexusSolverParameter:
        """Sets the solver parameters for the SOLVER keyword in the NexusSolverParameter object."""
        next_value = fo.get_next_value(0, file_as_list=[line])
        next_value = next_value.upper() if next_value is not None else None
        if next_value in SOLVER_SCOPE_KEYWORDS:
            current_solver_scope = next_value
            # get the value of the next token and assign it to the solver_parameter_for_timestep object
            solver_scoped_keyword = fo.get_expected_token_value(next_value, line, file_list=self.file_content)
            attribute_value, type_assignment = self.__get_solver_attribute_value(
                current_solver_scope, solver_scoped_keyword)
            value = fo.get_token_value(solver_scoped_keyword, line, file_list=[line])
            if value is not None:
                solver_parameter_for_timestep.__setattr__(attribute_value, type_assignment(value))
            else:
                # cover for the fact that it may not have a next value
                solver_parameter_for_timestep.__setattr__(attribute_value,
                                                          type_assignment(solver_scoped_keyword))

        elif next_value in ['ITERATIVE', 'DIRECT']:
            attribute_value, type_assignment = self.__get_solver_attribute_value(
                current_solver_scope, next_value)
            solver_parameter_for_timestep.__setattr__(attribute_value, type_assignment(next_value))

        elif next_value in SOLVER_SCOPED_KEYWORDS:
            # if the first value we find is in the SOLVER_SCOPED_KEYWORDS, then we have a keyword that is
            # not a token, but a value referring to a scope. So we can directly get the value and assign it
            # to the solver_parameter_for_timestep object.
            attribute_value, type_assignment = self.__get_solver_attribute_value(
                current_solver_scope, next_value)
            value = fo.get_token_value(next_value, line, file_list=self.file_content)
            solver_parameter_for_timestep.__setattr__(attribute_value, type_assignment(value))

        elif next_value in ['NOCUT', 'CUT']:
            solver_parameter_for_timestep.solver_timestep_cut = next_value == 'CUT'
        elif next_value in ['PRECON_ILU', 'PRECON_AMG', 'PRECON_AMG_RS']:
            solver_parameter_for_timestep.solver_precon = next_value
            precon_setting = fo.get_token_value(next_value, line, file_list=[line])
            if precon_setting is not None:
                precon_value_num = fo.get_expected_token_value(precon_setting, line,
                                                               file_list=self.file_content)
                solver_parameter_for_timestep.solver_precon_value = float(precon_value_num)
                solver_parameter_for_timestep.solver_precon_setting = precon_setting
        elif next_value in SOLVER_KEYWORDS:
            solver_parameter_for_timestep = self.__get_solver_token_values(next_value,
                                                                           line, solver_parameter_for_timestep)
        return solver_parameter_for_timestep

    def __get_solver_token_values(self, solver_token_value: str, line: str,
                                  solver_parameter_for_timestep: NexusSolverParameter) -> NexusSolverParameter:
        """Get the values for the SOLVER KEYWORD VALUE format."""
        solver_token_value = solver_token_value.upper()
        keyword_mapping = NexusSolverParameter.solver_keyword_mapping()

        # get the attribute value and type assignment
        if solver_token_value in SOLVER_SCOPE_KEYWORDS:
            property_set = fo.get_expected_token_value(solver_token_value, line, file_list=[line])
            attribute_value, type_assignment = self.__get_solver_attribute_value(solver_token_value, property_set)
            value = fo.get_expected_token_value(property_set, line, file_list=self.file_content)

        else:
            attribute_value, type_assignment = keyword_mapping[solver_token_value]
            value = fo.get_expected_token_value(solver_token_value, line, file_list=self.file_content)
        if type_assignment is bool:
            # set to True if the value is 'ON' and False if the value is 'OFF' (the bool conversion happens
            # by type_assignment)
            value = '' if value.upper() == 'OFF' else value
        solver_parameter_for_timestep.__setattr__(attribute_value, type_assignment(value))
        return solver_parameter_for_timestep

    @staticmethod
    def __get_solver_attribute_value(solver_scope: str, solver_scoped_keyword: str) -> tuple[str, type]:
        """Get the attribute value and type assignment for a solver keyword and a following value like CYCLELENGTH that
        has several mappings.
        """
        keyword_mapping = NexusSolverParameter.solver_keyword_mapping()
        lookup_attribute_value = solver_scope.upper() + ' ' + solver_scoped_keyword.upper()
        attribute_value, type_assignment = keyword_mapping[lookup_attribute_value]

        return attribute_value, type_assignment

    def __get_impstab_token_values(self, token_value: str, line: str,
                                   solver_parameter_for_timestep: NexusSolverParameter) -> NexusSolverParameter:
        keyword_mapping = NexusSolverParameter.impstab_keyword_mapping()
        attribute_value, type_assignment = keyword_mapping[token_value.upper()]

        match token_value:
            case 'USEMASSCFL':
                value = ''
            case 'COATS':
                value = 'COATS'
            case 'PEACEMAN':
                value = 'PEACEMAN'
            case _:
                value = fo.get_expected_token_value(token_value, line, file_list=self.file_content)

        solver_parameter_for_timestep.__setattr__(attribute_value, type_assignment(value))
        return solver_parameter_for_timestep

    def __get_generic_solver_token_values(self, next_token: str, line: str,
                                          solver_parameter_for_timestep: NexusSolverParameter,
                                          current_solver_param_token: str
                                          ) -> NexusSolverParameter:
        """Get the values for the generic KEYWORD VALUE format when following a token like DT, TOLS, GRIDSOLVER, etc."""
        # get the keyword mapping for the current solver parameter section
        if current_solver_param_token in MAX_CHANGE_KEYWORDS:
            attribute_value, type_assignment = (
                NexusSolverParameter.get_max_change_attribute_name(keyword=current_solver_param_token,
                                                                   value=next_token))
        else:
            keyword_mapping = self.__get_keyword_mapping_for_current_solver_param(current_solver_param_token)
            # get the ressimpy attribute name
            attribute_value, type_assignment = keyword_mapping[next_token.upper()]

        value = fo.get_expected_token_value(next_token, line, file_list=self.file_content)
        # add the value to the solver_parameter_for_timestep object
        solver_parameter_for_timestep.__setattr__(attribute_value, type_assignment(value))
        return solver_parameter_for_timestep

    def __get_keyword_mapping_for_current_solver_param(self, current_solver_param_token: str) -> (
            dict)[str, tuple[str, type]]:
        """Get the keyword mapping for the current solver parameter section."""
        solver_param_to_keyword_mapping = {'DT': NexusSolverParameter.dt_keyword_mapping(),
                                           'SOLVER': NexusSolverParameter.solver_keyword_mapping(),
                                           'IMPSTAB': NexusSolverParameter.impstab_keyword_mapping(),
                                           'GRIDSOLVER': NexusSolverParameter.gridsolver_keyword_mapping(),
                                           'SOLO': NexusSolverParameter.solo_keyword_mapping(),
                                           'TOLS': NexusSolverParameter.tols_keyword_mapping(),
                                           }
        keyword_mapping = solver_param_to_keyword_mapping[current_solver_param_token]
        return keyword_mapping

    def to_string_for_date(self, date: ISODateTime) -> str:
        """Outputs the string of all the solver parameters for a given date.

        Args:
        date (ISODateTime): Date for which to output the solver parameters in ISODateTime format.

        Returns:
            str: String representation of the solver parameters for the given date.
        """

        solver_parameters_for_date = [x for x in self.solver_parameters if x.iso_date == date]
        output_str = ''

        for solver_param in solver_parameters_for_date:
            output_str += solver_param.to_string()
        return output_str



"""Class to hold Nexus Rock properties."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import numpy as np
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.rock_keywords import ROCK_ALL_TABLE_KEYWORDS, ROCK_KEYWORDS_VALUE_FLOAT
from ResSimpy.Nexus.NexusKeywords.rock_keywords import ROCK_SINGLE_KEYWORDS, ROCK_KEYWORDS_VALUE_STR
from ResSimpy.Nexus.NexusKeywords.rock_keywords import ROCK_KEYWORDS, ROCK_REV_IRREV_OPTIONS
from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import RockUnits
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo
import ResSimpy.FileOperations.file_operations as fo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusRockMethod(DynamicProperty):
    """Class to hold Nexus Rock properties.

    Attributes:
        file (NexusFile): Nexus rock properties file object
        input_number (int): Rock properties method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific rock properties method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                dict[str, Union[float, pd.DataFrame]]]] = \
        field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem
    __hash__ = DynamicProperty.__hash__

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        """Initialises the NexusRockMethod class.

        Args:
            file (NexusFile): Nexus file containing the rock method.
            input_number (int): method number for the rock method
            model_unit_system (UnitSystem): unit system from the model
            properties: dictionary of the properties for the rock method. Defaults to None.
        """
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'CR': ('rock_compressibility', float),
            'KP': ('rock_permeability_compressibility', float),
            'PREF': ('reference_pressure', float),
            'P': ('pressure', float),
            'DP': ('delta_pressure', float),
        }
        return keywords

    @property
    def units(self) -> RockUnits:
        """Returns the attribute to unit map for the rock method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return RockUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with rock properties data, in Nexus file format."""
        printable_str = ''
        rock_dict = self.properties
        for key, value in rock_dict.items():
            if key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif isinstance(value, pd.DataFrame):
                printable_str += f'{key}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n\n'
            elif isinstance(value, dict):
                printable_str += f"{key.replace('_', ' ')}\n"
                for subkey in value.keys():
                    printable_str += f"SWINIT {subkey}\n"
                    df = value[subkey]
                    if isinstance(df, pd.DataFrame):
                        printable_str += df.to_string(na_rep='', index=False) + '\n'
                    printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif value == '':
                printable_str += f'{key}\n'
            else:
                printable_str += f'{key} {value}\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus rock properties file contents and populate NexusRockMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize properties
        cmt_indices: list[int] = []
        wirct_indices_dict: dict[str, dict[str, list[int]]] = {}
        # Flag to tell when to start reading a table
        start_reading_table: bool = False
        swinit_key = ''
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {}
        for table_name in ROCK_ALL_TABLE_KEYWORDS:
            table_being_read[table_name] = False

        line_indx = 0
        for line in file_as_list:

            # Find ROCK key-value pairs, such as PREF 2000 or CR 1e-6
            if [i for i in line.split() if i in ROCK_KEYWORDS_VALUE_FLOAT]:
                for key in ROCK_KEYWORDS_VALUE_FLOAT:
                    if nfo.check_token(key, line):
                        self.properties[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            # Find standalone rock property keywords, such as COMPR or KPMULT
            if [i for i in line.split() if i in ROCK_SINGLE_KEYWORDS]:
                for word in ROCK_SINGLE_KEYWORDS:
                    if nfo.check_token(word, line):
                        self.properties[word] = ''
            # Handle REVERSIBLE or IRREVERSIBLE keywords
            if [i for i in line.split() if i in ROCK_KEYWORDS_VALUE_STR]:
                for key in ROCK_KEYWORDS_VALUE_STR:
                    if nfo.check_token(key, line):
                        if fo.get_token_value(key, line, file_as_list) in ROCK_REV_IRREV_OPTIONS:
                            self.properties[key] = nfo.get_expected_token_value(key, line, file_as_list)
                        else:
                            self.properties[key] = ''
            # Find starting index of rock compaction table
            if nfo.check_token('CMT', line):
                cmt_indices = [line_indx + 1, len(file_as_list)]
                table_being_read['CMT'] = True
                start_reading_table = True
                line_indx += 1
                continue
            # Find ending index of rock compaction table
            if start_reading_table and table_being_read['CMT']:
                for potential_endkeyword in ROCK_KEYWORDS:
                    if nfo.check_token(potential_endkeyword, line):
                        cmt_indices[1] = line_indx
                        start_reading_table = False
                        table_being_read['CMT'] = False
                        break
            # Find ending index of a water-induced rock compaction table
            if start_reading_table and table_being_read['WIRCT']:
                for potential_endkeyword in ROCK_KEYWORDS:
                    if nfo.check_token(potential_endkeyword, line):
                        wirct_indices_dict['WIRCT'][swinit_key][1] = line_indx
                        start_reading_table = False
                        table_being_read['WIRCT'] = False
                        break
            # Find starting index of a water-induced rock compaction table
            if nfo.check_token('WIRCT', line):
                wirct_indices_dict['WIRCT'] = {}
            if nfo.check_token('SWINIT', line):
                swinit_key = nfo.get_expected_token_value('SWINIT', line, file_as_list)
                wirct_indices_dict['WIRCT'][swinit_key] = [line_indx+1, len(file_as_list)]
                table_being_read['WIRCT'] = True
                start_reading_table = True

            line_indx += 1

        # Read in compaction table(s) if there are any
        if len(cmt_indices) > 0:
            self.properties['CMT'] = nfo.read_table_to_df(file_as_list[cmt_indices[0]:cmt_indices[1]])
        for key in wirct_indices_dict.keys():
            self.properties[key] = {}
            for subkey in wirct_indices_dict[key].keys():
                property_key = self.properties[key]
                if not isinstance(property_key, dict):
                    raise ValueError(f"Property is not a dictionary: {str(self.properties[key])}")
                property_key[subkey] = nfo.read_table_to_df(file_as_list[
                                                            wirct_indices_dict[key][subkey][0]:
                                                            wirct_indices_dict[key][subkey][1]])



"""Class for handling targets in the Nexus Network.

This class is used to store and manipulate the targets in a NexusNetwork. It is stored as an instance in the
NexusNetwork class as "targets".
"""
from __future__ import annotations
from dataclasses import dataclass, field
from uuid import UUID
from typing import Sequence, Optional, TYPE_CHECKING

import pandas as pd

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.FileOperations.File import File
from ResSimpy.Nexus.DataModels.NexusOptions import NexusOptions
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.Nexus.nexus_collect_tables import collect_all_tables_to_objects
from ResSimpy.Nexus.DataModels.Network.NexusTarget import NexusTarget
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.nexus_modify_object_in_file import ModifyObjectOperations
from ResSimpy.Nexus.nexus_remove_object_from_file import RemoveObjectOperations
from ResSimpy.GenericContainerClasses.Targets import Targets
from ResSimpy.Utils.obj_to_dataframe import obj_to_dataframe

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass(kw_only=True)
class NexusTargets(Targets):
    """Class for handling targets in the Nexus Network.

    This class is used to store and manipulate the targets in a NexusNetwork. It is stored as an instance in the
    NexusNetwork class as "targets".
    """
    _targets: list[NexusTarget] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        """Initialises the NexusTargets class.

        Args:
            parent_network (NexusNetwork): The network that the targets are a part of.
        """
        self.__parent_network: NexusNetwork = parent_network
        self._targets: list[NexusTarget] = []
        self.__add_object_operations = AddObjectOperations(None, self.table_header,
                                                           self.table_footer, self.__parent_network.model)
        self.__remove_object_operations = RemoveObjectOperations(self.__parent_network, self.table_header,
                                                                 self.table_footer)
        self.__modify_object_operations = ModifyObjectOperations(self)

    @property
    def table_header(self) -> str:
        """Start of the target definition table."""
        return 'TARGET'

    @property
    def table_footer(self) -> str:
        """End of the target definition table."""
        return 'ENDTARGET'

    def get_all(self) -> Sequence[NexusTarget]:
        """Returns a list of targets loaded from the simulator."""
        self.__parent_network.get_load_status()
        return self._targets

    def get_by_name(self, target_name: str) -> Optional[NexusTarget]:
        """Returns a single target with the provided name loaded from the simulator.

        Args:
        ----
            target_name (str): name of the requested node

        Returns:
        -------
            NexusTarget: which has the same name as the requested target_name

        """
        targets_to_return = filter(lambda x: False if x.name is None else x.name.upper() == target_name.upper(),
                                   self._targets)
        return next(targets_to_return, None)

    def get_df(self) -> pd.DataFrame:
        """Creates a dataframe representing all processed target data in a surface file.

        Returns:
            DataFrame: of the properties of the targets through time with each row representing a target.
        """
        self.__parent_network.get_load_status()
        df_store = obj_to_dataframe(self._targets)
        return df_store

    def get_overview(self) -> str:
        """Returns overview of the Targets."""
        raise NotImplementedError('To be implemented')

    def load(self, surface_file: File, start_date: str, default_units: UnitSystem) -> None:
        """Calls load target and appends the list of discovered nodes into the NexusTargets object.

        Args:
            surface_file (File): NexusFile representation of the surface file.
            start_date (str): Starting date of the run
            default_units (UnitSystem): Units used in case not specified by surface file.

        Raises:
            TypeError: if the unit system found in the property check is not a valid enum UnitSystem.

        """
        new_targets, _ = collect_all_tables_to_objects(surface_file, {'TARGET': NexusTarget},
                                                       start_date=start_date,
                                                       default_units=default_units,
                                                       date_format=self.__parent_network.model.date_format)
        cons_list = new_targets.get('TARGET')
        self._add_to_memory(cons_list)

    def _add_to_memory(self, additional_list: Optional[list[NexusTarget]]) -> None:
        """Extends the targets object by a list of targets provided to it.

        Args:
        ----
            additional_list (Sequence[NexusTarget]): list of nexus targets to add to the targets list.

        Returns:
        -------
            None
        """
        if additional_list is None:
            return
        self._targets.extend(additional_list)

    def remove(self, target_to_remove: DataObjectMixinDictType | UUID) -> None:
        """Remove a node from the network based on the properties matching a dictionary or id.

        Args:
            target_to_remove (UUID | DataObjectMixinDictType): UUID of the node to remove or a dictionary \
            with sufficient matching parameters to uniquely identify a node

        """
        self.__remove_object_operations.remove_object_from_network_main(
            target_to_remove, self._network_element_name, self._targets)

    def add(self, target_to_add: DataObjectMixinDictType) -> None:
        """Adds a target to a network, taking a dictionary with properties for the new node.

        Args:
            target_to_add (DataObjectMixinDictType): dictionary taking all the properties for the new \
            target. Requires date and a target name.
        """
        new_object = self.__add_object_operations.add_network_obj(target_to_add, NexusTarget, self.__parent_network)
        self._add_to_memory([new_object])

    def modify(self, target_to_modify: DataObjectMixinDictType,
               new_properties: DataObjectMixinDictType) -> None:
        """Modifies an existing node based on a matching dictionary of properties.

        Partial matches allowed if precisely 1 matching node is found. Updates the properties with properties in the
        new_properties dictionary.

        Args:
            target_to_modify (DataObjectMixinDictType): dictionary containing attributes to match in the
            existing node set.
            new_properties (DataObjectMixinDictType): properties to switch to in the new node
        """
        self.__parent_network.get_load_status()

        self.__modify_object_operations.modify_network_object(target_to_modify, new_properties,
                                                              self.__parent_network)

    def _look_up_region_numbers_for_targets(self, options_file: NexusOptions) -> None:
        """Looks up the region numbers for the targets based on the region names in the options file.

        Args:
            options_file (NexusOptions): options file object

        """
        for target in self._targets:
            if target.region is not None:
                target.region_number = options_file.look_up_region_number_by_name(target.region)



"""Contains a class for representing a set of ConList objects for the Nexus model."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, Sequence

from ResSimpy.DataModelBaseClasses.NetworkList import NetworkList
from ResSimpy.GenericContainerClasses.NetworkLists import NetworkLists
from ResSimpy.Nexus.DataModels.Network.NexusConList import NexusConList

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass(kw_only=True)
class NexusConLists(NetworkLists):
    """Class for representing a set of conList objects for the Nexus model."""

    _lists: list[NexusConList] = field(default_factory=list)
    __parent_network: NexusNetwork

    def __init__(self, parent_network: NexusNetwork, con_lists: None | list[NexusConList] = None) -> None:
        """Initialises the NexusConLists class. Supports ADD and REMOVE operations on the conlist.
        Currently, this does not handle the keywords: NODEIN, NODEOUT, NODEPRODWH, NODEPRODBH, NODEINJWH or NODEINJBH.

        Args:
            parent_network (NexusNetwork): The network that the con lists are a part of.
            con_lists (list[NexusConList]): The list of con lists to be added to the NexusConLists object
        """
        self._lists = con_lists if con_lists is not None else []
        self.__parent_network = parent_network
        super().__init__(parent_network=parent_network)

    def get_all(self) -> list[NexusConList]:
        """Returns all conList names."""
        self.__parent_network.get_load_status()
        return self._lists

    @property
    def conlists(self) -> list[NexusConList]:
        """Returns list of all conList objects."""
        self.__parent_network.get_load_status()
        return self._lists

    @staticmethod
    def table_header() -> Literal['CONLIST']:
        """Start of the Node definition table."""
        return 'CONLIST'

    @staticmethod
    def table_footer() -> Literal['ENDCONLIST']:
        """End of the Node definition table."""
        return 'ENDCONLIST'

    @property
    def _network_element_name(self) -> Literal['conlists']:
        return 'conlists'

    def _add_to_memory(self, additional_list: None | Sequence[NetworkList]) -> None:
        """Extends the conlist stored in memory by a list of conlist provided to it."""
        if additional_list is None:
            return
        if not isinstance(additional_list, list):
            additional_list = list(additional_list)
        only_conlist = [x for x in additional_list if isinstance(x, NexusConList)]
        self._lists.extend(only_conlist)



from dataclasses import dataclass

from ResSimpy.DataModelBaseClasses.GridArrayDefinition import GridArrayDefinition
from ResSimpy.DataModelBaseClasses.LGR import LGR


@dataclass
class NexusLGR(LGR):
    """Class for handling a Local Grid Refinement (LGR) in the NexusGrid."""
    _dx: GridArrayDefinition
    _dy: GridArrayDefinition
    _dz: GridArrayDefinition
    _depth: GridArrayDefinition
    _mdepth: GridArrayDefinition
    _dznet: GridArrayDefinition
    _compr: GridArrayDefinition
    _icoars: GridArrayDefinition
    _ialphaf: GridArrayDefinition
    _ipolymer: GridArrayDefinition
    _iadsorption: GridArrayDefinition
    _itracer: GridArrayDefinition
    _igrid: GridArrayDefinition
    _isector: GridArrayDefinition
    _swl: GridArrayDefinition
    _swr: GridArrayDefinition
    _swu: GridArrayDefinition
    _sgl: GridArrayDefinition
    _sgr: GridArrayDefinition
    _sgu: GridArrayDefinition
    _swro: GridArrayDefinition
    _swro_ls: GridArrayDefinition
    _sgro: GridArrayDefinition
    _sgrw: GridArrayDefinition
    _krw_swro: GridArrayDefinition
    _krws_ls: GridArrayDefinition
    _krw_swu: GridArrayDefinition
    _krg_sgro: GridArrayDefinition
    _krg_sgu: GridArrayDefinition
    _krg_sgrw: GridArrayDefinition
    _kro_swl: GridArrayDefinition
    _kro_swr: GridArrayDefinition
    _kro_sgl: GridArrayDefinition
    _kro_sgr: GridArrayDefinition
    _krw_sgl: GridArrayDefinition
    _krw_sgr: GridArrayDefinition
    _sgtr: GridArrayDefinition
    _sotr: GridArrayDefinition
    _swlpc: GridArrayDefinition
    _sglpc: GridArrayDefinition
    _pcw_swl: GridArrayDefinition
    _pcg_sgu: GridArrayDefinition
    _chloride: GridArrayDefinition
    _calcium: GridArrayDefinition
    _salinity: GridArrayDefinition
    _api: GridArrayDefinition
    _tmx: GridArrayDefinition
    _tmy: GridArrayDefinition
    _tmz: GridArrayDefinition
    _multbv: GridArrayDefinition
    _pv: GridArrayDefinition
    _pvmult: GridArrayDefinition
    _livecell: GridArrayDefinition
    _worka1: GridArrayDefinition
    _worka2: GridArrayDefinition
    _worka3: GridArrayDefinition
    _worka4: GridArrayDefinition
    _worka5: GridArrayDefinition
    _worka6: GridArrayDefinition
    _worka7: GridArrayDefinition
    _worka8: GridArrayDefinition
    _worka9: GridArrayDefinition
    _kxeff: GridArrayDefinition
    _kyeff: GridArrayDefinition
    _kzeff: GridArrayDefinition

    def __init__(self, parent_grid: str, name: str, i1: int, i2: int, j1: int, j2: int, k1: int, k2: int,
                 nx: list[int], ny: list[int], nz: list[int]) -> None:
        """Initializes the NexusLGR class."""
        super().__init__(parent_grid, name, i1, i2, j1, j2, k1, k2, nx, ny, nz)
        self._pvmult: GridArrayDefinition = GridArrayDefinition()
        self._livecell: GridArrayDefinition = GridArrayDefinition()
        self._worka1: GridArrayDefinition = GridArrayDefinition()
        self._worka2: GridArrayDefinition = GridArrayDefinition()
        self._worka3: GridArrayDefinition = GridArrayDefinition()
        self._worka4: GridArrayDefinition = GridArrayDefinition()
        self._worka5: GridArrayDefinition = GridArrayDefinition()
        self._worka6: GridArrayDefinition = GridArrayDefinition()
        self._worka7: GridArrayDefinition = GridArrayDefinition()
        self._worka8: GridArrayDefinition = GridArrayDefinition()
        self._worka9: GridArrayDefinition = GridArrayDefinition()
        self._dx: GridArrayDefinition = GridArrayDefinition()
        self._dy: GridArrayDefinition = GridArrayDefinition()
        self._dz: GridArrayDefinition = GridArrayDefinition()
        self._depth: GridArrayDefinition = GridArrayDefinition()
        self._mdepth: GridArrayDefinition = GridArrayDefinition()
        self._dznet: GridArrayDefinition = GridArrayDefinition()
        self._compr: GridArrayDefinition = GridArrayDefinition()
        self._icoars: GridArrayDefinition = GridArrayDefinition()
        self._ialphaf: GridArrayDefinition = GridArrayDefinition()
        self._ipolymer: GridArrayDefinition = GridArrayDefinition()
        self._iadsorption: GridArrayDefinition = GridArrayDefinition()
        self._itracer: GridArrayDefinition = GridArrayDefinition()
        self._igrid: GridArrayDefinition = GridArrayDefinition()
        self._isector: GridArrayDefinition = GridArrayDefinition()
        self._swl: GridArrayDefinition = GridArrayDefinition()
        self._swr: GridArrayDefinition = GridArrayDefinition()
        self._swu: GridArrayDefinition = GridArrayDefinition()
        self._sgl: GridArrayDefinition = GridArrayDefinition()
        self._sgr: GridArrayDefinition = GridArrayDefinition()
        self._sgu: GridArrayDefinition = GridArrayDefinition()
        self._swro: GridArrayDefinition = GridArrayDefinition()
        self._swro_ls: GridArrayDefinition = GridArrayDefinition()
        self._sgro: GridArrayDefinition = GridArrayDefinition()
        self._sgrw: GridArrayDefinition = GridArrayDefinition()
        self._krw_swro: GridArrayDefinition = GridArrayDefinition()
        self._krws_ls: GridArrayDefinition = GridArrayDefinition()
        self._krw_swu: GridArrayDefinition = GridArrayDefinition()
        self._krg_sgro: GridArrayDefinition = GridArrayDefinition()
        self._krg_sgu: GridArrayDefinition = GridArrayDefinition()
        self._krg_sgrw: GridArrayDefinition = GridArrayDefinition()
        self._kro_swl: GridArrayDefinition = GridArrayDefinition()
        self._kro_swr: GridArrayDefinition = GridArrayDefinition()
        self._kro_sgl: GridArrayDefinition = GridArrayDefinition()
        self._kro_sgr: GridArrayDefinition = GridArrayDefinition()
        self._krw_sgl: GridArrayDefinition = GridArrayDefinition()
        self._krw_sgr: GridArrayDefinition = GridArrayDefinition()
        self._sgtr: GridArrayDefinition = GridArrayDefinition()
        self._sotr: GridArrayDefinition = GridArrayDefinition()
        self._swlpc: GridArrayDefinition = GridArrayDefinition()
        self._sglpc: GridArrayDefinition = GridArrayDefinition()
        self._pcw_swl: GridArrayDefinition = GridArrayDefinition()
        self._pcg_sgu: GridArrayDefinition = GridArrayDefinition()
        self._chloride: GridArrayDefinition = GridArrayDefinition()
        self._calcium: GridArrayDefinition = GridArrayDefinition()
        self._salinity: GridArrayDefinition = GridArrayDefinition()
        self._api: GridArrayDefinition = GridArrayDefinition()
        self._tmx: GridArrayDefinition = GridArrayDefinition()
        self._tmy: GridArrayDefinition = GridArrayDefinition()
        self._tmz: GridArrayDefinition = GridArrayDefinition()
        self._multbv: GridArrayDefinition = GridArrayDefinition()
        self._pv: GridArrayDefinition = GridArrayDefinition()
        self._kxeff: GridArrayDefinition = GridArrayDefinition()
        self._kyeff: GridArrayDefinition = GridArrayDefinition()
        self._kzeff: GridArrayDefinition = GridArrayDefinition()

    @property
    def netgrs(self) -> GridArrayDefinition:
        """Returns the netgrs values."""
        return self._netgrs

    @property
    def porosity(self) -> GridArrayDefinition:
        """Returns the porosity values."""
        return self._porosity

    @property
    def sw(self) -> GridArrayDefinition:
        """Returns the sw values."""
        return self._sw

    @property
    def sg(self) -> GridArrayDefinition:
        """Returns the sg values."""
        return self._sg

    @property
    def pressure(self) -> GridArrayDefinition:
        """Returns the pressure values."""
        return self._pressure

    @property
    def temperature(self) -> GridArrayDefinition:
        """Returns the temperature values."""
        return self._temperature

    @property
    def kx(self) -> GridArrayDefinition:
        """Returns the kx values."""
        return self._kx

    @property
    def ky(self) -> GridArrayDefinition:
        """Returns the ky values."""
        return self._ky

    @property
    def kz(self) -> GridArrayDefinition:
        """Returns the kz values."""
        return self._kz

    @property
    def dx(self) -> GridArrayDefinition:
        """Returns the dx values."""
        return self._dx

    @property
    def dy(self) -> GridArrayDefinition:
        """Returns the dy values."""
        return self._dy

    @property
    def dz(self) -> GridArrayDefinition:
        """Returns the dz values."""
        return self._dz

    @property
    def depth(self) -> GridArrayDefinition:
        """Returns the depth values."""
        return self._depth

    @property
    def mdepth(self) -> GridArrayDefinition:
        """Returns the mdepth values."""
        return self._mdepth

    @property
    def dznet(self) -> GridArrayDefinition:
        """Returns the dznet values."""
        return self._dznet

    @property
    def compr(self) -> GridArrayDefinition:
        """Returns the compr values."""
        return self._compr

    @property
    def icoars(self) -> GridArrayDefinition:
        """Returns the icoars values."""
        return self._icoars

    @property
    def ialphaf(self) -> GridArrayDefinition:
        """Returns the ialphaf values."""
        return self._ialphaf

    @property
    def ipolymer(self) -> GridArrayDefinition:
        """Returns the ipolymer values."""
        return self._ipolymer

    @property
    def iadsorption(self) -> GridArrayDefinition:
        """Returns the iadsorption values."""
        return self._iadsorption

    @property
    def itracer(self) -> GridArrayDefinition:
        """Returns the itracer values."""
        return self._itracer

    @property
    def igrid(self) -> GridArrayDefinition:
        """Returns the igrid values."""
        return self._igrid

    @property
    def isector(self) -> GridArrayDefinition:
        """Returns the isector values."""
        return self._isector

    @property
    def swl(self) -> GridArrayDefinition:
        """Returns the swl values."""
        return self._swl

    @property
    def swr(self) -> GridArrayDefinition:
        """Returns the swr values."""
        return self._swr

    @property
    def swu(self) -> GridArrayDefinition:
        """Returns the swu values."""
        return self._swu

    @property
    def sgl(self) -> GridArrayDefinition:
        """Returns the sgl values."""
        return self._sgl

    @property
    def sgr(self) -> GridArrayDefinition:
        """Returns the sgr values."""
        return self._sgr

    @property
    def sgu(self) -> GridArrayDefinition:
        """Returns the sgu values."""
        return self._sgu

    @property
    def swro(self) -> GridArrayDefinition:
        """Returns the swro values."""
        return self._swro

    @property
    def swro_ls(self) -> GridArrayDefinition:
        """Returns the swro_ls values."""
        return self._swro_ls

    @property
    def sgro(self) -> GridArrayDefinition:
        """Returns the sgro values."""
        return self._sgro

    @property
    def sgrw(self) -> GridArrayDefinition:
        """Returns the sgrw values."""
        return self._sgrw

    @property
    def krw_swro(self) -> GridArrayDefinition:
        """Returns the krw_swro values."""
        return self._krw_swro

    @property
    def krws_ls(self) -> GridArrayDefinition:
        """Returns the krws_ls values."""
        return self._krws_ls

    @property
    def krw_swu(self) -> GridArrayDefinition:
        """Returns the krw_swu values."""
        return self._krw_swu

    @property
    def krg_sgro(self) -> GridArrayDefinition:
        """Returns the krg_sgro values."""
        return self._krg_sgro

    @property
    def krg_sgu(self) -> GridArrayDefinition:
        """Returns the krg_sgu values."""
        return self._krg_sgu

    @property
    def krg_sgrw(self) -> GridArrayDefinition:
        """Returns the krg_sgrw values."""
        return self._krg_sgrw

    @property
    def kro_swl(self) -> GridArrayDefinition:
        """Returns the kro_swl values."""
        return self._kro_swl

    @property
    def kro_swr(self) -> GridArrayDefinition:
        """Returns the kro_swr values."""
        return self._kro_swr

    @property
    def kro_sgl(self) -> GridArrayDefinition:
        """Returns the kro_sgl values."""
        return self._kro_sgl

    @property
    def kro_sgr(self) -> GridArrayDefinition:
        """Returns the kro_sgr values."""
        return self._kro_sgr

    @property
    def krw_sgl(self) -> GridArrayDefinition:
        """Returns the krw_sgl values."""
        return self._krw_sgl

    @property
    def krw_sgr(self) -> GridArrayDefinition:
        """Returns the krw_sgr values."""
        return self._krw_sgr

    @property
    def sgtr(self) -> GridArrayDefinition:
        """Returns the sgtr values."""
        return self._sgtr

    @property
    def sotr(self) -> GridArrayDefinition:
        """Returns the sotr values."""
        return self._sotr

    @property
    def swlpc(self) -> GridArrayDefinition:
        """Returns the swlpc values."""
        return self._swlpc

    @property
    def sglpc(self) -> GridArrayDefinition:
        """Returns the sglpc values."""
        return self._sglpc

    @property
    def pcw_swl(self) -> GridArrayDefinition:
        """Returns the pcw_swl values."""
        return self._pcw_swl

    @property
    def pcg_sgu(self) -> GridArrayDefinition:
        """Returns the pcg_sgu values."""
        return self._pcg_sgu

    @property
    def chloride(self) -> GridArrayDefinition:
        """Returns the chloride values."""
        return self._chloride

    @property
    def calcium(self) -> GridArrayDefinition:
        """Returns the calcium values."""
        return self._calcium

    @property
    def salinity(self) -> GridArrayDefinition:
        """Returns the salinity values."""
        return self._salinity

    @property
    def api(self) -> GridArrayDefinition:
        """Returns the api values."""
        return self._api

    @property
    def tmx(self) -> GridArrayDefinition:
        """Returns the tmx values."""
        return self._tmx

    @property
    def tmy(self) -> GridArrayDefinition:
        """Returns the tmy values."""
        return self._tmy

    @property
    def tmz(self) -> GridArrayDefinition:
        """Returns the tmz values."""
        return self._tmz

    @property
    def multbv(self) -> GridArrayDefinition:
        """Returns the multbv values."""
        return self._multbv

    @property
    def pv(self) -> GridArrayDefinition:
        """Returns the pv values."""
        return self._pv

    @property
    def iequil(self) -> GridArrayDefinition:
        """Returns the iequil values."""
        return self._iequil

    @property
    def ipvt(self) -> GridArrayDefinition:
        """Returns the ipvt values."""
        return self._ipvt

    @property
    def iwater(self) -> GridArrayDefinition:
        """Returns the iwater values."""
        return self._iwater

    @property
    def irelpm(self) -> GridArrayDefinition:
        """Returns the irelpm values."""
        return self._irelpm

    @property
    def irock(self) -> GridArrayDefinition:
        """Returns the irock values."""
        return self._irock

    @property
    def itran(self) -> GridArrayDefinition:
        """Returns the itran values."""
        return self._itran

    @property
    def iregion(self) -> dict[str, GridArrayDefinition]:
        """Returns the iregion values."""
        return self._iregion

    @property
    def pvmult(self) -> GridArrayDefinition:
        """Returns the pvmult values."""
        return self._pvmult

    @property
    def livecell(self) -> GridArrayDefinition:
        """Returns the livecell values."""
        return self._livecell

    @property
    def worka1(self) -> GridArrayDefinition:
        """Returns the worka1 values."""
        return self._worka1

    @property
    def worka2(self) -> GridArrayDefinition:
        """Returns the worka2 values."""
        return self._worka2

    @property
    def worka3(self) -> GridArrayDefinition:
        """Returns the worka3 values."""
        return self._worka3

    @property
    def worka4(self) -> GridArrayDefinition:
        """Returns the worka4 values."""
        return self._worka4

    @property
    def worka5(self) -> GridArrayDefinition:
        """Returns the worka5 values."""
        return self._worka5

    @property
    def worka6(self) -> GridArrayDefinition:
        """Returns the worka6 values."""
        return self._worka6

    @property
    def worka7(self) -> GridArrayDefinition:
        """Returns the worka7 values."""
        return self._worka7

    @property
    def worka8(self) -> GridArrayDefinition:
        """Returns the worka8 values."""
        return self._worka8

    @property
    def worka9(self) -> GridArrayDefinition:
        """Returns the worka9 values."""
        return self._worka9

    @property
    def kxeff(self) -> GridArrayDefinition:
        """Returns the kxeff values."""
        return self._kxeff

    @property
    def kyeff(self) -> GridArrayDefinition:
        """Returns the kyeff values."""
        return self._kyeff

    @property
    def kzeff(self) -> GridArrayDefinition:
        """Returns the kzeff values."""
        return self._kzeff



from abc import ABC
from dataclasses import dataclass

from ResSimpy.DataModelBaseClasses.GridArrayDefinition import GridArrayDefinition


@dataclass
class LGR(ABC):
    """Abstract base class for Local Grid Refinements (LGR)."""
    _parent_grid: str
    _name: str
    # range of grid blocks
    _i1: int
    _i2: int
    _j1: int
    _j2: int
    _k1: int
    _k2: int
    _nx: list[int]
    _ny: list[int]
    _nz: list[int]

    _netgrs: GridArrayDefinition
    _porosity: GridArrayDefinition
    _sw: GridArrayDefinition
    _sg: GridArrayDefinition
    _pressure: GridArrayDefinition
    _temperature: GridArrayDefinition
    _kx: GridArrayDefinition
    _ky: GridArrayDefinition
    _kz: GridArrayDefinition

    _iequil: GridArrayDefinition
    _ipvt: GridArrayDefinition
    _iwater: GridArrayDefinition
    _irelpm: GridArrayDefinition
    _irock: GridArrayDefinition
    _itran: GridArrayDefinition
    _iregion: dict[str, GridArrayDefinition]

    def __init__(self, parent_grid: str, name: str, i1: int, i2: int, j1: int, j2: int, k1: int, k2: int,  # noqa: D417
                 nx: list[int], ny: list[int], nz: list[int]) -> None:
        """Initializes the LGR class.

        Args:
            parent_grid (str): Name of the parent grid.
            name (str): Name of the LGR.
            i1, i2, j1, j2, k1, k2 (int): Range of grid blocks.
            nx, ny, nz (list[int]): number of x/y/z direction fine gridblocks for each corresponding coarse gridblock.
        """
        self._parent_grid = parent_grid
        self._name = name

        self._i1 = i1
        self._i2 = i2
        self._j1 = j1
        self._j2 = j2
        self._k1 = k1
        self._k2 = k2

        self._nx = nx
        self._ny = ny
        self._nz = nz

        self._netgrs = GridArrayDefinition()
        self._porosity = GridArrayDefinition()
        self._sw = GridArrayDefinition()
        self._sg = GridArrayDefinition()
        self._pressure = GridArrayDefinition()
        self._temperature = GridArrayDefinition()
        self._kx = GridArrayDefinition()
        self._ky = GridArrayDefinition()
        self._kz = GridArrayDefinition()

        self._iequil: GridArrayDefinition = GridArrayDefinition()
        self._ipvt: GridArrayDefinition = GridArrayDefinition()
        self._iwater: GridArrayDefinition = GridArrayDefinition()
        self._irelpm: GridArrayDefinition = GridArrayDefinition()
        self._irock: GridArrayDefinition = GridArrayDefinition()
        self._itran: GridArrayDefinition = GridArrayDefinition()
        self._iregion: dict[str, GridArrayDefinition] = {}

    @property
    def range(self) -> tuple[int, int, int, int, int, int]:
        """Returns the range of grid blocks in the parent grid."""
        return (self._i1, self._i2, self._j1, self._j2, self._k1, self._k2)

    @property
    def name(self) -> str:
        """Returns the name of the LGR."""
        return self._name

    @property
    def parent_grid(self) -> str:
        """Returns the name of the parent grid."""
        return self._parent_grid

    @property
    def nx(self) -> list[int]:
        """Returns the x-direction refinement values."""
        return self._nx

    @property
    def ny(self) -> list[int]:
        """Returns the y-direction refinement values."""
        return self._ny

    @property
    def nz(self) -> list[int]:
        """Returns the z-direction refinement values."""
        return self._nz

    @property
    def netgrs(self) -> GridArrayDefinition:
        """Returns the net-to-gross grid array definition."""
        return self._netgrs

    @property
    def porosity(self) -> GridArrayDefinition:
        """Returns the porosity grid array definition."""
        return self._porosity

    @property
    def sw(self) -> GridArrayDefinition:
        """Returns the water saturation grid array definition."""
        return self._sw

    @property
    def sg(self) -> GridArrayDefinition:
        """Returns the gas saturation grid array definition."""
        return self._sg

    @property
    def pressure(self) -> GridArrayDefinition:
        """Returns the pressure grid array definition."""
        return self._pressure

    @property
    def temperature(self) -> GridArrayDefinition:
        """Returns the temperature grid array definition."""
        return self._temperature

    @property
    def kx(self) -> GridArrayDefinition:
        """Returns the kx grid array definition."""
        return self._kx

    @property
    def ky(self) -> GridArrayDefinition:
        """Returns the ky grid array definition."""
        return self._ky

    @property
    def kz(self) -> GridArrayDefinition:
        """Returns the kz grid array definition."""
        return self._kz

    @property
    def iequil(self) -> GridArrayDefinition:
        """Returns the iequil grid array definition."""
        return self._iequil

    @property
    def ipvt(self) -> GridArrayDefinition:
        """Returns the ipvt grid array definition."""
        return self._ipvt

    @property
    def iwater(self) -> GridArrayDefinition:
        """Returns the iwater grid array definition."""
        return self._iwater

    @property
    def irelpm(self) -> GridArrayDefinition:
        """Returns the irelpm grid array definition."""
        return self._irelpm

    @property
    def irock(self) -> GridArrayDefinition:
        """Returns the irock grid array definition."""
        return self._irock

    @property
    def itran(self) -> GridArrayDefinition:
        """Returns the itran grid array definition."""
        return self._itran

    @property
    def iregion(self) -> dict[str, GridArrayDefinition]:
        """Returns the iregion grid array definition."""
        return self._iregion

    @property
    def grid_array_defs(self) -> dict[str, GridArrayDefinition]:
        """Returns the grid array definitions in a dictionary keyed by the ResSimpy keyword associated with the
        grid array definition.
        """
        grid_array_defs = {'netgrs': self._netgrs,
                           'porosity': self._porosity,
                           'sw': self._sw,
                           'sg': self._sg,
                           'pressure': self._pressure,
                           'temperature': self._temperature,
                           'kx': self._kx,
                           'ky': self._ky,
                           'kz': self._kz,
                           'iequil': self._iequil,
                           'ipvt': self._ipvt,
                           'iwater': self._iwater,
                           'irelpm': self._irelpm,
                           'irock': self._irock,
                           'itran': self._itran}
        # filter to only include grid array definitions with a modifier
        grid_array_defs = {key: value for key, value in grid_array_defs.items() if value.modifier is not None or
                           value.value is not None or value.mods is not None}
        return grid_array_defs



"""Class for collecting all the Nexus Aquifer methods."""
from __future__ import annotations
from dataclasses import dataclass, field
import os
from typing import Optional, MutableMapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusAquiferMethod import NexusAquiferMethod
from ResSimpy.DataModelBaseClasses.Aquifer import Aquifer
from ResSimpy.Utils.dynamic_method_manipulations import add_dynamic_method


@dataclass(kw_only=True)
class NexusAquiferMethods(Aquifer):
    """Class for collection of Nexus aquifer methods.

    Attributes:
        inputs (dict[int, NexusAquiferMethod]): Collection of Nexus aquifer methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of aquifer files, as defined in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusAquiferMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem
    _model_files: Optional[FcsNexusFile] = field(default=None, repr=False, compare=False)

    def __init__(self, model_unit_system: UnitSystem, inputs: Optional[MutableMapping[int, NexusAquiferMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None, assume_loaded: bool = False,
                 model_files: Optional[FcsNexusFile] = None) -> None:
        """Initialises the NexusAquiferMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusAquiferMethod]]): Collection of Nexus aquifer methods.
            files (Optional[dict[int, NexusFile]]): Collection of aquifer files, as defined in Nexus fcs file.
            Keyed by the method number.
            assume_loaded (bool): If True, assumes that the properties are already loaded.
            model_files (Optional[FcsNexusFile]): The Nexus fcs file that contains the whole set of model files.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusAquiferMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        self.__properties_loaded = assume_loaded
        self._model_files = model_files
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing aquifer methods."""
        if not self.__properties_loaded:
            self.load_aquifer_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'AQUIFER method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusAquiferMethod]:
        """Returns mapping of aquifer where keys are of type int.
        If properties are not loaded, it will load them.
        """
        if not self.__properties_loaded:
            self.load_aquifer_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        """Returns dictionary of Nexus files where keys are of type int."""
        return self.__files

    def load_aquifer_methods(self) -> None:
        """Loads a collection of aquifer method files.
        This method checks if aquifer files are available and read their properties into
        the 'NexusAquifer' object.
        """
        # Read in aquifer properties from Nexus aquifer method files
        if self.__files is not None and len(self.__files) > 0:  # Check if aquifer files exist
            for table_num in self.__files.keys():  # For each aquifer property method
                aquifer_file = self.__files[table_num]
                if aquifer_file.location is None:
                    raise ValueError(f'Unable to find aquifer file: {aquifer_file.location}')
                if os.path.isfile(aquifer_file.location):
                    # Create NexusAquifer object
                    self.__inputs[table_num] = NexusAquiferMethod(file=aquifer_file, input_number=table_num,
                                                                  model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with aquifer properties from file
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system

    @property
    def keyword(self) -> str:
        """Return the keyword for the aquifer methods."""
        return 'AQUIFER'

    def add_method(self, method: DynamicProperty, new_file_name: str, create_new_file: bool = False) -> None:
        """Adds a new aquifer method to the collection.

        Args:
            method (DynamicProperty): The aquifer method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        add_dynamic_method(dynamic_method_collection=self, method=method, new_file_name=new_file_name,
                           create_new_file=create_new_file)

    def _method_type(self) -> type[NexusAquiferMethod]:
        """Returns the expected type of the dynamic property."""
        return NexusAquiferMethod



from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin
from dataclasses import dataclass
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping
import pandas as pd
from pandas import DataFrame


@dataclass
class NexusIprMethod(DataObjectMixin):
    """Class to hold IPR Method."""

    __table: pd.DataFrame

    def __init__(self, date: str, table: pd.DataFrame) -> None:
        """Individual IPRTable class to add read of IPR files.

        Args:
             date(str): IPR table date.
             table(pd.DataFrame): IPR table.
        """

        self.__table = table

        super().__init__(date=date)

    @property
    def units(self) -> BaseUnitMapping:
        """Returns BaseUnitMapping."""
        return BaseUnitMapping(unit_system=None)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type[float]]]:
        """Gets the mapping of keywords to attribute definitions."""
        mapping = {
            'PRES': ('pressure', float),
            'QO': ('oil_surface_rate', float),
            'QW': ('water_surface_rate', float),
            'QG': ('gas_surface_rate', float),
        }
        return mapping

    @property
    def table(self) -> DataFrame:
        """Returns table as DataFrame."""
        return self.__table



"""Class for collection of Nexus rock property methods."""
from __future__ import annotations
from dataclasses import dataclass, field
import os
from typing import Optional, MutableMapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.FcsFile import FcsNexusFile
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusRockMethod import NexusRockMethod
from ResSimpy.DataModelBaseClasses.Rock import Rock
from ResSimpy.Utils.dynamic_method_manipulations import add_dynamic_method


@dataclass(kw_only=True)
class NexusRockMethods(Rock):
    """Class for collection of Nexus rock property methods.

    Attributes:
        inputs (dict[int, NexusRockMethod]): Collection of Nexus rock property methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of rock property files, as defined in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusRockMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem
    _model_files: Optional[FcsNexusFile] = field(default=None, repr=False, compare=False)

    def __init__(self, model_unit_system: UnitSystem,
                 inputs: Optional[MutableMapping[int, NexusRockMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None, assume_loaded: bool = False,
                 model_files: Optional[FcsNexusFile] = None) -> None:
        """Initialises the NexusRockMethods class.

        Args:
            model_unit_system (UnitSystem): Unit system used in the model.
            inputs (Optional[MutableMapping[int, NexusRockMethod]]): Collection of Nexus rock property methods.
            files (Optional[dict[int, NexusFile]]): Collection of rock property files, as defined in Nexus fcs file.
                Keyed by the method number.
            assume_loaded (bool): If True, assumes that the rock methods have already been loaded.
            model_files (Optional[FcsNexusFile]): The FcsFile object that contains all the model files for keeping the
                files in sync across the whole model.
        """
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusRockMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        self.__properties_loaded = assume_loaded
        self._model_files = model_files
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing rock methods."""
        if not self.__properties_loaded:
            self.load_rock_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'ROCK method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusRockMethod]:
        """Returns mapping of Nexus rock properties.
        If properties have not been loaded, it loads them first.
        """
        if not self.__properties_loaded:
            self.load_rock_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        """Returns dictionary of NexusFile where keys are of type int."""
        return self.__files

    def load_rock_methods(self) -> None:
        """Loads rock property files from Nexus fcs file."""
        # Read in rock properties from Nexus rock method files
        if self.__files is not None and len(self.__files) > 0:  # Check if rock files exist
            for table_num in self.__files.keys():  # For each rock property method
                rock_file = self.__files[table_num]
                if rock_file.location is None:
                    raise ValueError(f'Unable to find rock file: {rock_file}')
                if os.path.isfile(rock_file.location):
                    # Create NexusRockMethod object
                    self.__inputs[table_num] = NexusRockMethod(file=rock_file, input_number=table_num,
                                                               model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with rock properties in file
        self.__properties_loaded = True

    @property
    def model_unit_system(self) -> UnitSystem:
        """Return the model unit system."""
        return self.__model_unit_system

    @property
    def keyword(self) -> str:
        """Returns the keyword for the rock methods."""
        return 'ROCK'

    def add_method(self, method: DynamicProperty, new_file_name: str, create_new_file: bool = False) -> None:
        """Adds a new rock method to the collection.

        Args:
            method (DynamicProperty): The rock method to add.
            new_file_name (str): The name of the file to save the method to.
            create_new_file (bool): Whether to create a new file for the method.
        """
        add_dynamic_method(dynamic_method_collection=self, method=method, new_file_name=new_file_name,
                           create_new_file=create_new_file)

    def _method_type(self) -> type[NexusRockMethod]:
        """Returns the expected type of the dynamic property."""
        return NexusRockMethod



from dataclasses import dataclass
from typing import Mapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.DataModelBaseClasses.DynamicPropertyContainer import DynamicPropertyContainer


@dataclass(kw_only=True)
class Water(DynamicPropertyContainer):
    """The abstract base class for a collection of water property inputs.

    Attributes:
        inputs (dict[int, DynamicProperty]): Collection of water property methods, as a dictionary.
    """

    @property
    def inputs(self) -> Mapping[int, DynamicProperty]:
        """A Collection of water property methods, as a dictionary."""
        raise NotImplementedError("Implement this in the derived class")



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Sequence, TYPE_CHECKING

from ResSimpy.Nexus.DataModels.Network.NexusAction import NexusAction

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass(kw_only=True)
class NexusActions:
    """A class representing a list of nexus action objects."""
    __actions: list[NexusAction] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        """Initialises the NexusActions class.

        Args:
            parent_network (NexusNetwork): The network that the actions are a part of.
            actions (list[NexusAction]): A list of nexus action objects.
        """
        self.__parent_network: NexusNetwork = parent_network
        self.__actions = []

    def get_all(self) -> Sequence[NexusAction]:
        """Returns a list of actions loaded from the simulator."""
        self.__parent_network.get_load_status()
        return self.__actions

    def _add_to_memory(self, actions_to_add: list[NexusAction]) -> None:
        """Adds the list of Nexus Action objects to memory."""
        self.__actions.extend(actions_to_add)
